# Copyright (c) 2010 Anil Kumar
# All rights reserved.
#
# License: BSD 

import os
import sys
import re

backend_plugins = []
backend_dict = {}

def _load_plugins(module, directory):
	pluginImports = __import__(module, globals(), locals())
	print('Scanning for backend plugins...')
	plist = []
	pdict = {}
	for i in sorted(os.listdir(directory)):
		path = os.path.join( directory, i, '__init__.py' )
		if os.path.isfile( path ):
			p = __import__( '%s.%s' % (module, i), globals(), locals(), ['*'] )
			plist.append(p)
			pdict[p.name()] = p
			if not hasattr(p, 'priority'):
				p.priority = 0 
	plist = sorted(plist, key=lambda p: p.priority, reverse=True)
	for p in plist:
		print('\t', p.name())
	return (plist, pdict)

def load_plugins():
	global backend_plugins, backend_dict
	(backend_plugins, backend_dict) = _load_plugins('backend.plugins', 'backend/plugins')

def plugin_list():
	return backend_plugins


prj = None

def proj_new_open_app_cb():
	prj.prj_feature_setup()

def proj_close_app_cb():
	pass

	

def _proj_new_open():
	proj_new_open_app_cb()

def proj_new(bname, proj_args):
	b = backend_dict[bname]

	global prj
	assert not prj
	prj = b.project_class().prj_new(proj_args)

	if prj:
		_proj_new_open()
	return prj != None

def plugin_guess(proj_path):
	bi = []
	for p in backend_plugins:
		if p.is_your_prj(proj_path):
			bi.append(p.name())
	return bi

def proj_open(proj_path, proj_type):
	b = backend_dict[proj_type]

	global prj
	prj = b.project_class().prj_open(proj_path)

	if prj:
		_proj_new_open()
	return prj != None

def proj_close():
	global prj
	prj.prj_close()
	prj = None
	
	from .plugins import CtagsCache
	CtagsCache.flush()

	proj_close_app_cb

def proj_is_open():
	return prj != None

def proj_name():
	return prj.prj_name() if prj else None

def proj_dir():
	return prj.prj_dir() if prj else None

def proj_src_files():
	return prj.prj_src_files()

def proj_conf():
	return prj.prj_conf()

def proj_settings_get():
	return prj.prj_settings_get()

def proj_settings_update(proj_args):
	return prj.prj_settings_update(proj_args)

def proj_is_ready():
	return prj.prj_is_ready()

def proj_query(rquery):
	return prj.prj_query(rquery)

def proj_rebuild():
	return prj.prj_rebuild()

def proj_query_fl():
	return prj.prj_query_fl()

def proj_type():
	return prj.prj_type()

def proj_feature():
	return prj.prj_feature()



#!/usr/bin/env python3

# Copyright (c) 2010 Anil Kumar
# All rights reserved.
#
# License: BSD 

from PyQt6.QtWidgets import *
from PyQt6.QtGui import *
from PyQt6.QtCore import *

class GenericResItem(QTreeWidgetItem):
	def __init__(self, li):
		QTreeWidgetItem.__init__(self, li)
	def column_val(self, col):
		return str(self.data(col, Qt.ItemDataRole.DisplayRole).toString())
	def line_val(self):
		return (int(self.column_val(1)))

class GenericRes(QTreeWidget):
	def __init__(self, parent=None):
		QTreeWidget.__init__(self)
		
		self.setColumnCount(1)
		self.setHeaderLabels(['Result'])

		#self.setMinimumHeight(200)
		self.setMinimumWidth(50)

		self.setFont(QFont("San Serif", 8))
		
		self.setIndentation(-2)

		self.setAllColumnsShowFocus(True)

	def add_result(self, res):
		for line in res:
			item = GenericResItem([line])
			self.addTopLevelItem(item)

class GenericFileCmdPage(QWidget):
	sig_goto_line = pyqtSignal(int)

	def __init__(self, parent=None):
		QWidget.__init__(self)
		self.ct = GenericRes(self)
		vlay = QVBoxLayout()
		#self.lbl = QLabel()
		#vlay.addWidget(self.lbl)
		vlay.addWidget(self.ct)
		self.setLayout(vlay)

		self.ct.itemActivated.connect(self.ct_itemActivated)

	def ct_itemActivated(self, item):
		try:
			line = int(str(item.data(1, Qt.ItemDataRole.DisplayRole).toString()))
		except:
			return
		self.sig_goto_line.emit(line)

	@staticmethod
	def run_plugin(cmd, filename, parent):
		if '%f' in cmd:
			args = cmd.replace('%f', filename)
			args = args.strip().split()
		else:
			args = cmd.strip().split()
			args.append(filename)
		import subprocess
		try:
			proc = subprocess.Popen(args, stdout=subprocess.PIPE)
			(res, err_data) = proc.communicate()
			res = res.decode()
		except Exception as e:
			res = '%s\n%s' % (' '.join(args), str(e))
		import re
		res = [ x.strip() for x in re.split(r'\r?\n', res.strip()) ]
		
		page = GenericFileCmdPage(parent)
		page.cmd = cmd
		#page.lbl.setText(' '.join(args))
		page.ct.add_result(res)
		parent.add_page(page, cmd)

def cmd_name():
	cmd_list = [
		#'stat',
		#'ls %f',
		#'ls %f %f'
	]
	import os
	cmd = os.getenv('SEASCOPE_FC_CUSTOM_CMD')
	if cmd:
		cmd_list += [ x.strip() for x in cmd.split(';') ]
	return list(set(cmd_list))

def run(filename, parent, cmd=None):
	if cmd:
		cmd_list = [ cmd ]
	else:
		cmd_list = cmd_name()
	for cmd in cmd_list:
		GenericFileCmdPage.run_plugin(cmd, filename, parent)



#!/usr/bin/env python3

# Copyright (c) 2010 Anil Kumar
# All rights reserved.
#
# License: BSD 

from PyQt6.QtWidgets import *
from PyQt6.QtGui import *
from PyQt6.QtCore import *

from . import CtagsManager

# Search Result
class SR:
	NONE  = 0
	LSKIP = 1
	LUPD  = 2
	MATCH = 3
	RUPD  = 4
	RSKIP = 5
	def __init__(self, k):
		self.line = k
		self.left = 0
		self.right = 1000000
		self.item = None
		self.done = False

class CtagsListItem(QTreeWidgetItem):
	def __init__(self, li):
		QTreeWidgetItem.__init__(self, li)
		if li[2] in [ 'class', 'interface', 'struct', 'type', 'enum', 'constrainer', '' ]:
			self.set_bold()
		if li[2] in [ 'reactor' ]:
			brush = self.foreground(0)
			brush.setColor(Qt.GlobalColor.gray)
			self.setForeground(0, brush)

	def set_bold(self):
		f = self.font(0)
		f.setBold(True)
		self.setFont(0, f)
	def column_val(self, col):
		return str(self.data(col, Qt.ItemDataRole.DisplayRole))
	def line_val(self):
		return (int(self.column_val(1)))

class CtagsList(QTreeWidget):
	def __init__(self, parent=None, isTree=False):
		QTreeWidget.__init__(self)
		
		self.setColumnCount(3)
		self.setHeaderLabels(['Name', 'Line', 'Type'])

		#self.setMinimumHeight(200)
		self.setMinimumWidth(50)

		self.setFont(QFont("San Serif", 8))
		
		if isTree:
			self.setIndentation(12)
			self.setExpandsOnDoubleClick(False)
			#self.setAnimated(True)
		else:
			self.setIndentation(-2)

		self.setAllColumnsShowFocus(True)

	def add_ct_result(self, res):
		for line in res:
			item = CtagsListItem(line)
			self.addTopLevelItem(item)
		#self.sortItems(1, Qt.SortOrder.AscendingOrder)
		#self.resizeColumnsToContents(1)
		self.resizeColumnToContents(0)
		self.resizeColumnToContents(1)
		self.resizeColumnToContents(2)

	def recurseTreeAdd(self, t, p):
		for k, v in list(t.items()):
			if k == '+':
				continue
			if k == '*':
				for line in v:
					item = CtagsListItem(line)
					p.addChild(item)
				continue
			if '+' in v:
				item = CtagsListItem(v['+'])
			else:
				item = CtagsListItem([k, '', '', ''])
			p.addChild(item)
			self.recurseTreeAdd(v, item)
		p.setExpanded(True)

	def add_ctree_result(self, res):
		p = self.invisibleRootItem()
		self.recurseTreeAdd(res, p)

		#self.sortItems(1, Qt.SortOrder.AscendingOrder)
		#self.resizeColumnsToContents(1)
		self.resizeColumnToContents(0)
		self.resizeColumnToContents(1)
		self.resizeColumnToContents(2)

	def ed_cursor_changed(self, line, pos):
		line = line + 1
		item = self.currentItem()
		if not item:
			item = self.topLevelItem(0)
			if not item:
				return
		if (item.line_val() == line):
			pass
		elif (item.line_val() < line):
			while True:
				next_item = self.itemBelow(item)
				if not next_item:
					break
				if (next_item.line_val() > line):
					break
				item = next_item
		else:
			while True:
				prev_item = self.itemAbove(item)
				if not prev_item:
					break
				item = prev_item
				if (item.line_val() <= line):
					break
		self.setCurrentItem(item)

	def search_item_check(self, item, sr):
		try:
			val = item.line_val()
		except:
			return SR.NONE
		if val < sr.left:
			return SR.LSKIP
		if val < sr.line:
			sr.left = val
			sr.item = item
			return SR.LUPD
		if val == sr.line:
			sr.left = val
			sr.item = item
			sr.done = True
			return SR.MATCH
		if val < sr.right:
			sr.right = val
			if not sr.item:
				sr.item = item
			return SR.RUPD
		return SR.RSKIP

	def search_recursive(self, p, sr):
		inx = 0;
		r = SR.NONE
		while inx < p.childCount():
			item = p.child(inx)
			if not item.childCount():
				break
			inx += 1
			if r < SR.MATCH:
				r = self.search_item_check(item, sr)
				if r == SR.MATCH:
					return
			self.search_recursive(item, sr)
			if sr.done:
				return
		if not inx < p.childCount():
			return

		last_item = p.child(p.childCount() - 1)
		r = self.search_item_check(last_item, sr)
		if r <= sr.MATCH:
			return

		r = SR.NONE
		while inx < p.childCount():
			item = p.child(inx)
			inx += 1
			r = self.search_item_check(item, sr)
			if r >= SR.MATCH:
				return

	def search_tree(self, line, pos):
		line = line + 1
		sr = SR(line)
		p = self.invisibleRootItem()
		self.search_recursive(p, sr)
		self.setCurrentItem(sr.item)
		

class CtagsListPage(QWidget):
	sig_goto_line = pyqtSignal(int)

	def __init__(self, parent=None):
		QWidget.__init__(self)
		self.le = QLineEdit()
		self.ct = CtagsList(self)
		vlay = QVBoxLayout()
		vlay.addWidget(self.le)
		vlay.addWidget(self.ct)
		self.setLayout(vlay)

		self.le.textChanged.connect(self.le_textChanged)
		self.le.returnPressed.connect(self.le_returnPressed)
		self.ct.itemActivated.connect(self.ct_itemActivated)

	def le_textChanged(self, text):
		#if (len(str(text)) == 1):
			#self.ct.keyboardSearch('')
		self.ct.keyboardSearch('')
		self.ct.keyboardSearch(text)

	def le_returnPressed(self):
		items = self.ct.selectedItems()
		if len(items) == 0:
			return
		self.ct.itemActivated.emit(items[0], 0)

	def ct_itemActivated(self, item):
		line_str = str(item.data(1, Qt.ItemDataRole.DisplayRole))
		try:
			line = int(line_str)
		except:
			return
		self.sig_goto_line.emit(line)
		self.le.clear()

	def keyPressEvent(self, ev):
		if ev.key() in [Qt.Key.Key_Up, Qt.Key.Key_Down, Qt.Key.Key_PageUp or Qt.Key.Key_PageDown]:
			self.ct.keyPressEvent(ev)
			return

	@staticmethod
	def do_ct_query(filename, parent):
		page = CtagsListPage(parent)
		res = CtagsManager.ct_query(filename)
		page.ct.add_ct_result(res)
		parent.add_page(page, 'C')
		parent.sig_ed_cursor_changed.connect(page.ct.ed_cursor_changed)

class CtagsTreePage(QWidget):
	sig_goto_line = pyqtSignal(int)

	def __init__(self, parent=None):
		QWidget.__init__(self)
		self.ct = CtagsList(self, isTree=True)
		vlay = QVBoxLayout()
		vlay.addWidget(self.ct)
		self.setLayout(vlay)

		self.ct.itemActivated.connect(self.ct_itemActivated)

	def ct_itemActivated(self, item):
		line_str = str(item.data(1, Qt.ItemDataRole.DisplayRole))
		try:
			line = int(line_str)
		except:
			return
		self.sig_goto_line.emit(line)

	@staticmethod
	def do_ct_query(filename, parent):
		(res, isTree) = CtagsManager.ct_tree_query(filename)
		if not isTree:
			return
		page = CtagsTreePage(parent)
		page.ct.add_ctree_result(res)
		parent.add_page(page, 'T')
		parent.sig_ed_cursor_changed.connect(page.ct.search_tree)

def run(filename, parent):
	CtagsListPage.do_ct_query(filename, parent)
	CtagsTreePage.do_ct_query(filename, parent)



#!/usr/bin/env python3

# Copyright (c) 2010 Anil Kumar
# All rights reserved.
#
# License: BSD 

from PyQt6.QtWidgets import *
from PyQt6.QtGui import *
from PyQt6.QtCore import *

from . import filecontext

class FileContextView(QTabWidget):
	sig_goto_line = pyqtSignal(int)
	sig_ed_cursor_changed = pyqtSignal(int, int)

	def __init__(self, parent=None):
		QTabWidget.__init__(self)
		self.setTabPosition(QTabWidget.TabPosition.South)

	def add_page(self, page, title):
		page.sig_goto_line.connect(self.sig_goto_line)
		self.addTab(page, title)
		self.setCurrentWidget(page)

	def run(self, filename):
		self.filename = filename
		filecontext.run_plugins(filename, self)

	def rerun(self, filename):
		inx = self.currentIndex()
		self.clear()
		self.run(filename)
		self.setCurrentIndex(inx)

	def focus_search_ctags(self):
		for inx in range(self.count()):
			page = self.widget(inx)
			if hasattr(page, 'le') and hasattr(page.le, 'setFocus'):
				self.setCurrentWidget(page)
				page.le.setFocus()
				break

	def get_already_opened_cmd_list(self):
		a_cmd_list = []
		for inx in range(self.count()):
			page = self.widget(inx)
			if hasattr(page, 'cmd'):
				a_cmd_list.append(page.cmd)
		return a_cmd_list

	def get_plugin_cmd_list(self):
		a_cmd_list = self.get_already_opened_cmd_list()
		cmd_list = []
		for p in filecontext.fc_plugins:
			if not hasattr(p, 'cmd_name'):
				continue
			cmd_name = p.cmd_name()
			if not cmd_name:
				continue
			if not isinstance(cmd_name, list):
				if cmd_name == '':
					continue
				cmd_name = [ cmd_name ]
			for cmd in cmd_name:
				if cmd in a_cmd_list:
					continue
				cmd_list.append((cmd, p))
		return cmd_list

	def menu_act_triggered_cb(self, act):
		act.plugin.run_plugin(self.filename, self, cmd=act.cmd_name)

	def mousePressEvent(self, m_ev):
		cmd_list = self.get_plugin_cmd_list()
		if len(cmd_list) == 0:
			return
		if (m_ev.button() == Qt.MouseButton.RightButton):
			# setup popup menu
			pmenu = QMenu()
			pmenu.triggered.connect(self.menu_act_triggered_cb)
			for (cmd_name, p) in cmd_list:
				act = pmenu.addAction(cmd_name)
				act.plugin = p
				act.cmd_name = cmd_name
			pmenu.exec(QCursor.pos())
			pmenu = None



# Copyright (c) 2010 Anil Kumar
# All rights reserved.
#
# License: BSD 

import os

def name():
	return 'generic_cmd'

def run_plugin(filename, parent, cmd=None):
	from . import GenericView
	GenericView.run(filename, parent, cmd=cmd)

def description():
	d = 'generic cmd description'
	return d

def cmd_name():
	from . import GenericView
	return GenericView.cmd_name()

priority = 100



#!/usr/bin/env python3

# Copyright (c) 2010 Anil Kumar
# All rights reserved.
#
# License: BSD 

import os, string, re

from ..PluginBase import PluginFeatureBase, ProjectBase, ConfigBase, QueryBase
from ..PluginBase import PluginProcess
from ..CtagsCache import CtagsThread


class GtagsFeature(PluginFeatureBase):
	def __init__(self):
		PluginFeatureBase.__init__(self)

		self.feat_desc = [
			['REF',	'-r'],
			['DEF',	''],
			#['<--',	'2'],
			['-->',	'-r'],
			#[	['TXT',	'4'],
			['GREP','-g'],
			['FIL',	'-P'],
			['INC',	'-g'],

			['QDEF', ''],
			['CTREE','12'],

			['CLGRAPH', '13'],
			['CLGRAPHD', '14'],
			['FFGRAPH', '14'],

			['UPD', '25'],
		]

		self.ctree_query_args = [
			['-->',	'--> F', 'Calling tree'			],
			#['<--',	'F -->', 'Called tree'			],
			['REF',	'==> F', 'Advanced calling tree'	],
		]
		
	def query_dlg_cb(self, req, cmd_str, in_opt):
		if req != '' and in_opt['substring']:
			req = '.*' + req + '.*'
		opt = None
		if in_opt['ignorecase']:
			opt = '-i'
		res = (cmd_str, req, opt)
		return res

class ConfigGtags(ConfigBase):
	def __init__(self):
		ConfigBase.__init__(self, 'gtags')

class ProjectGtags(ProjectBase):
	def __init__(self):
		ProjectBase.__init__(self)

	@staticmethod
	def _prj_new_or_open(conf):
		prj = ProjectGtags()
		prj.feat = GtagsFeature()
		prj.conf = conf
		prj.qry = QueryGtags(prj.conf, prj.feat)

		return (prj)

	@staticmethod
	def prj_new(proj_args):
		d = proj_args[0]
		prj = ProjectGtags.prj_open(d)
		return None

	@staticmethod
	def prj_open(proj_path):
		conf = ConfigGtags()
		conf.proj_open(proj_path)
		prj = ProjectGtags._prj_new_or_open(conf)
		return (prj)


class GtProcess(PluginProcess):
	def __init__(self, wdir, rq):
		PluginProcess.__init__(self, wdir, rq)
		self.name = 'gtags process'

	def parse_result(self, text, sig):
		text = re.split(r'\r?\n', text)
		if self.cmd_str == 'FIL':
			res = [ ['',  line.split(' ')[0], '', '' ] for line in text if line != '' ]
			return res
		res = []
		for line in text:
			if line == '':
				continue
			line = line.split(' ', 3)
			line = ['<unknown>', line[0], line[2], line[3]]
			res.append(line)

		CtagsThread(sig).apply_fix(self.cmd_str, res, ['<unknown>'])

		return None

class QueryGtags(QueryBase):
	def __init__(self, conf, feat):
		QueryBase.__init__(self)
		self.conf = conf
		self.feat = feat

	def query(self, rquery):
		if (not self.conf):
		#or not self.conf.is_ready()):
			print("pm_query not is_ready")
			return None
		cmd_str = rquery['cmd']
		req     = rquery['req']
		opt     = rquery['opt']
		if opt == None or opt == '':
			opt = []
		else:
			opt = opt.split()
		cmd_opt = self.feat.cmd_str2id[cmd_str]
		pargs = [ 'global', '-a', '--result=cscope', '-x' ] + opt
		if cmd_opt != '':
			pargs += [ cmd_opt ]
		pargs += [ '--', req ]
		
		qsig = GtProcess(self.conf.c_dir, [cmd_str, req]).run_query_process(pargs, req, rquery)
		return qsig

	def rebuild(self):
		if (not self.conf.is_ready()):
			print("pm_query not is_ready")
			return None
		if (os.path.exists(os.path.join(self.conf.c_dir, 'GTAGS'))):
			pargs = [ 'global', '-u' ]
		else:
			pargs = [ 'gtags', '-i' ]
		qsig = GtProcess(self.conf.c_dir, None).run_rebuild_process(pargs)
		return qsig

	def query_fl(self):
		if not os.path.exists(os.path.join(self.conf.c_dir, 'GTAGS')):
			return []
		pargs = [ 'global', '-P', '-a' ]
		qsig = GtProcess(self.conf.c_dir, None).run_query_fl(pargs)
		return qsig

	def gt_is_open(self):
		return self.conf != None
	def gt_is_ready(self):
		return self.conf.is_ready()



# Copyright (c) 2010 Anil Kumar
# All rights reserved.
#
# License: BSD 

import os
import sys
import re

backend_plugins = []
backend_dict = {}

def _load_plugins(module, directory):
	pluginImports = __import__(module, globals(), locals())
	print('Scanning for backend plugins...')
	plist = []
	pdict = {}
	for i in sorted(os.listdir(directory)):
		path = os.path.join( directory, i, '__init__.py' )
		if os.path.isfile( path ):
			p = __import__( '%s.%s' % (module, i), globals(), locals(), ['*'] )
			plist.append(p)
			pdict[p.name()] = p
			if not hasattr(p, 'priority'):
				p.priority = 0 
	plist = sorted(plist, key=lambda p: p.priority, reverse=True)
	for p in plist:
		print('\t', p.name())
	return (plist, pdict)

def load_plugins():
	global backend_plugins, backend_dict
	(backend_plugins, backend_dict) = _load_plugins('backend.plugins', 'backend/plugins')

def plugin_list():
	return backend_plugins

def plugin_guess(proj_path):
	bi = []
	for p in backend_plugins:
		if p.is_your_prj(proj_path):
			bi.append(p.name())
	return bi


class BProject:
	def __init__(self):
		self.prj = None

	def proj_new_open_app_cb(self):
		self.prj.prj_feature_setup()

	def proj_close_app_cb(self):
		pass

	def _proj_new_open(self):
		self.proj_new_open_app_cb()

	def proj_new(self, bname, proj_args):
		b = backend_dict[bname]

		assert not self.prj
		prj = b.project_class().prj_new(proj_args)

		if prj:
			_proj_new_open()
		return self.prj != None

	def proj_open(self, proj_path, proj_type):
		b = backend_dict[proj_type]

		self.prj = b.project_class().prj_open(proj_path)

		if self.prj:
			self._proj_new_open()
		return self.prj != None

	def proj_close(self):
		self.prj.prj_close()
		self.prj = None
		
		from .plugins import CtagsCache
		CtagsCache.flush()

		self.proj_close_app_cb()

	def proj_is_open(self):
		return self.prj != None

	def proj_name(self):
		return self.prj.prj_name() if self.prj else None

	def proj_dir(self):
		return self.prj.prj_dir() if self.prj else None

	def proj_src_files(self):
		return self.prj.prj_src_files()

	def proj_conf(self):
		return self.prj.prj_conf()

	def proj_settings_get(self):
		return self.prj.prj_settings_get()

	def proj_settings_update(proj_args):
		return self.prj.prj_settings_update(proj_args)

	def proj_is_ready(self):
		return self.prj.prj_is_ready()

	def proj_query(self, rquery):
		return self.prj.prj_query(rquery)

	def proj_rebuild(self):
		return self.prj.prj_rebuild()

	def proj_query_fl(self):
		return self.prj.prj_query_fl()

	def proj_type(self):
		return self.prj.prj_type()

	def proj_feature(self):
		return self.prj.prj_feature()



# Copyright (c) 2010 Anil Kumar
# All rights reserved.
#
# License: BSD

import os

def name():
	return 'idutils'

def is_your_prj(path):
	f = os.path.join(path, 'ID')
	return os.path.exists(f)

def project_class():
	from .IdutilsProject import ProjectIdutils
	return ProjectIdutils

def description():
	d = 'Preferred backend of seascope.\n'
	d += 'Some advanced features of seascope are available only with this backend\n'
	d += '\n'
	d += 'It\'s really "idutils + ctags" backend.\n'
	d += '    - idutils provides the indexing facility\n'
	d += '    - ctags provides the language intelligence\n'
	d += 'This backend combines both of these intelligently to provide superior source code browsing facility\n'
	d += '\n'
	d += 'A wide variety of languages are supported (ctags --list-languages)\n'
	d += 'For example C, C++, Java, Python etc\n'
	d += '\n'
	d += 'If your source files are not being indexed by idutils then configure it using /usr/share/id-lang.map\n'
	return d

priority = 800



# Copyright (c) 2010 Anil Kumar
# All rights reserved.
#
# License: BSD 

import os, re
import copy
import subprocess

def NOT_IMPLEMENTED(n, f):
	msg = '%s: %s: Not implemeted' % (n, f)
	print(msg)
	assert false

from . import CtagsCache

is_seascope_debug = os.getenv('SEASCOPE_DEBUG')

cmd_table_master = [
	[	'REF',		['&References',		'Ctrl+0'],	['References to'	]	],
	[	'DEF',		['&Definitions',	'Ctrl+1'],	['Definition of'	]	],
	[	'<--',		['&Called Functions',	'Ctrl+2'],	['Functions called by'	]	],
	[	'-->',		['C&alling Functions',	'Ctrl+3'],	['Functions calling'	]	],
	[	'TXT',		['Find &Text',		'Ctrl+4'],	['Find text'		]	],
	[	'GREP',		['Find &Egrep',		'Ctrl+5'],	['Find egrep pattern'	]	],
	[	'FIL',		['Find &File',		'Ctrl+7'],	['Find files'		]	],
	[	'INC',		['&Include/Import',	'Ctrl+8'],	['Find include/import'	]	],
	[	'---',		[None,				],	None				],
	[	'QDEF', 	['&Quick Definition',	'Ctrl+]'],	None				],
	[	'CTREE',	['Call Tr&ee',		'Ctrl+\\'],	['Call tree'		]	],
	[	'---',		[None,				],	None				],
	[	'CLGRAPH',  	['Class &Graph',	'Ctrl+:'],	['Class graph'		]	],
	[	'CLGRAPHD', 	['Class Graph Dir',	'Ctrl+;'],	['Class graph dir'	]	],
	[	'FFGRAPH', 	['File Func Graph',	'Ctrl+^'],	['File Func graph dir'	]	],
	[	'---',		[None,				],	None				],
	[	'UPD',		['Re&build Database',	None	],	None				],
]
class PluginFeatureBase:
	def __init__(self):
		self.clgraph_query_args = [
			['CLGRAPH',	'D', 'Derived classes'			],
			['CLGRAPH',	'B', 'Base classes'			],
		]
		self.clgraph_query_args = [
			['CLGRAPH',	'D', 'Derived classes'			],
			['CLGRAPH',	'B', 'Base classes'			],
		]
		self.ffgraph_query_args = [
			['FFGRAPH',    'F',   'File functions graph'],
			['FFGRAPH_E',  'F+E', 'File functions + external graph'],
			['FFGRAPH_D',  'D',   'Directory functions graph'],
			['FFGRAPH_DE', 'D+E', 'Directory functions + external graph']
		]
	
	def setup(self):
		feat_cmds = [ d[0] for d in self.feat_desc ]
		ct = copy.deepcopy(cmd_table_master)
		self.cmd_table = [ t for t in ct if t[0] in feat_cmds or t[0] == '---' ]

		self.menu_cmd_list = [ [c[0]] + c[1] for c in self.cmd_table ]

		self.cmd_str2id = {}
		self.cmd_str2qstr = {}
		self.cmd_qstr2str = {}
		for c in self.feat_desc:
			self.cmd_str2id[c[0]] = c[1]
		for c in self.cmd_table:
			if c[2] != None:
				self.cmd_str2qstr[c[0]] = c[2][0]
				self.cmd_qstr2str[c[2][0]] = c[0]
		# python 2.7
		#self.cmd_str2id = { c[0]:c[1] for c in self.feat_desc }
		#self.cmd_str2qstr = { c[0]:c[2][0] for c in self.cmd_table if c[1] }
		#self.cmd_qstr2str = { c[2][0]:c[0] for c in self.cmd_table if c[1] }
		self.cmd_qstrlist = [ c[2][0] for c in self.cmd_table if c[2] ]

class ProjectBase:
	prj = None
	qry = None

	def __init__(self):
		self.feat = None

	def prj_close(self):
		if (self.conf != None):
			self.conf.proj_close()
		self.conf = None

	def prj_dir(self):
		return self.conf.c_dir
	def prj_name(self):
		return self.conf.get_proj_name()
	def prj_src_files(self):
		return self.conf.get_proj_src_files()

	def prj_is_open(self):
		return self.conf != None
	def prj_is_ready(self):
		return self.conf.is_ready()

	def prj_conf(self):
		return self.conf.get_proj_conf()
		
	def prj_update_conf(self, proj_args):
		self.conf.proj_update(proj_args)

	def prj_show_settings(self, proj_args):
		NOT_IMPLEMENTED(__name__, __func__)
	def prj_settings(self, proj_args):
		NOT_IMPLEMENTED(__name__, __func__)

	def prj_feature_setup(self):
		self.feat.setup()

	def prj_query(self, rquery):
		return self.qry.query(rquery)

	def prj_rebuild(self):
		return self.qry.rebuild()

	def prj_query_fl(self):
		return self.qry.query_fl()

	def prj_type(self):
		return self.conf.prj_type

	def prj_feature(self):
		return self.feat

	def prj_settings_get(self):
		proj_args = self.prj_conf()
		return proj_args

	def prj_settings_update(self, proj_args):
		NOT_IMPLEMENTED(__name__, __func__)
		return

class ConfigBase:
	def __init__(self, ptype):
		self.prj_type = ptype
		self.c_dir = ''
		self.c_opt = ''
		self.c_flist = []

	def get_proj_name(self):
		return os.path.split(self.c_dir)[1]

	def get_proj_src_files(self):
		fl = self.c_flist
		return fl

	def get_proj_conf(self):
		return (self.c_dir, self.c_opt, self.c_flist)

	def read_config(self):
		pass
	def write_config(self):
		pass

	def proj_start(self):
		pass

	def proj_open(self, proj_path):
		self.c_dir = proj_path
		self.read_config()
		self.proj_start()

	def proj_update(self, proj_args):
		self.proj_new(proj_args)
		
	def proj_new(self, proj_args):
		self.proj_args = proj_args
		(self.c_dir, self.c_opt, self.c_flist) = proj_args
		self.write_config()
		self.proj_start()

	def proj_close(self):
		pass

	def is_ready(self):
		return True

	@staticmethod
	def prepare_menu(menubar):
		pass

class QueryBase:
	@staticmethod
	def prepare_menu(menubar):
		pass

	def __init__(self):
		pass

	def query(self, rquery):
		cmd_str = rquery['cmd']
		req     = rquery['req']
		opt     = rquery['opt']
		if opt == None:
			opt = []

		pargs = []
		if cmd_str == 'GREP':
			pargs = ['grep', '-E', '-R', '-n', '-I']
			pargs += [ '--', req ]
			pargs += [self.conf.c_dir]
		else:
			assert(false)
			return None
		qsig = PluginProcess(self.conf.c_dir, [cmd_str, req]).run_query_process(pargs, req, rquery)
		return qsig
		
	def rebuild():
		NOT_IMPLEMENTED(__name__, __func__)

	def conf_is_open(self):
		return self.conf != None
	def conf_is_ready(self):
		return self.conf.is_ready()
		

class QueryUiBase:
	def __init__(self):
		pass

class QuerySignal:
	def __init__(self):
		pass

	def _relevancy_sort(self, hfile, res):
		pt = []
		pd = {}
		p = hfile
		(pre, ext) = os.path.splitext(hfile)
		c = None
		while p != c:
			e = [p, [], []]
			pt += [e]
			pd[p] = e
			c = p
			p = os.path.dirname(p)
		for line in res:
			f = line[1]
			d = os.path.dirname(f)
			p = f
			while p not in pd:
				p = os.path.dirname(p)
			e = pd[p]
			if p in [f, d]:
				e[1].append(line)
			else:
				e[2].append(line)
		for e in pt:
			e[1] = sorted(e[1], key=lambda li: li[1])
			e[2] = sorted(e[2], key=lambda li: li[1])
		pre = pre + '.*'
		e0 = []
		e1 = []
		for e in pt[1][1]:
			if re.match(pre, e[1]):
				e0 += [e]
			else:
				e1 += [e]
		pt[0][1] += e0
		pt[1][1] = e1

		res1 = []
		res2 = []
		for e in pt:
			res1 += e[1]
			res2 += e[2]
		res = res1 + res2
		return res

	def relevancy_sort(self, res):
		if os.getenv('RELEVANCY_SORT', 1) == 0:
			return res
		hint_file = None
		try:
			hint_file = self.rquery['hint_file']
		except:
			pass
		if not hint_file:
			return res
		if not os.path.isabs(hint_file):
			print('BUG: relevancy_sort: not abs path:', hint_file)
			return res
		if len(res) > 10000:
			return res
		return self._relevancy_sort(hint_file, res)

class PluginProcessBase:
	proc_list = []

	def __init__(self, wdir):
		self.wdir = wdir
		self.proc = None
		self.outd = {}

		PluginProcess.proc_list.append(self)
		self.is_rebuild = False
		self.is_query_fl = False

		self.sig = QuerySignal()

	def _cleanup(self):
		PluginProcess.proc_list.remove(self)
		e = self.outd['err_data']
		if e:
			s = '<b>' + self.p_cmd + '</b><p>' + '<p>'.join(e.splitlines())
			print('s')
		r = self.outd['err_data']
		if r:
			s = '<b>' + self.p_cmd + '</b><p>Summary<p>' + r
			print('s')

	def run_process(self, pargs):
		if is_seascope_debug:
			print(self.p_cmd)
		try:
			self.proc = subprocess.Popen(pargs, cwd=self.wdir, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
			(out_data, err_data) = self.proc.communicate()
			print('ANIL2 %s' % out_data.decode())
			print('ANIL %s' % out_data)
			self.outd['out_data'] = out_data.decode()
			self.outd['err_data'] = err_data.decode()

			#rc = self.proc.returncode
			#if rc:
				#print 'PluginProcess.run(): returncode', rc
				#self.outd['out_data'] = None
				#self.outd['err_data'] = str(pargs), 'exited with returncode %s' % rc
		except Exception as e:
			print('PluginProcess.run():', e)
			self.outd['out_data'] = None
			self.outd['err_data'] = str(e)

		self._finished_cb()

	def run_query_process(self, pargs, sym, rquery=None):
		self.sig.sym = sym
		self.sig.rquery = rquery
		self.p_cmd = ' '.join(pargs)
		self.run_process(pargs)
		return self.outd

	def _finished_cb(self):
		if 'out_data' in self.outd:
			#print 'output', res
			#print 'cmd:', self.p_cmd

			res = self.outd['out_data']
			self.outd['out_data']

			if self.is_rebuild:
				pass
			elif self.is_query_fl:
				res = self.parse_query_fl(res)
			else:
				# XXX result_dbg
				try:
					res = self.parse_result(res, self.sig)
					res = self.sig.relevancy_sort(res);
				except Exception as e:
					print(e)
					res = [['', '', '', 'error while parsing output of: ' + self.p_cmd]]

			self.outd['res'] = res
		self._cleanup()

	def run_rebuild_process(self, pargs):
		self.is_rebuild = True
		self.p_cmd = ' '.join(pargs)
		self.run_process(pargs)
		CtagsCache.flush()
		return self.outd

	def run_query_fl(self, pargs):
		self.is_query_fl = True
		self.p_cmd = ' '.join(pargs)
		self.run_process(pargs)
		return self.outd

	def parse_query_fl(self, text):
		fl = []
		for f in re.split(r'\r?\n', text.strip()):
			if f == '':
				continue
			fl.append(os.path.join(self.wdir, f))
		return fl

class PluginProcess(PluginProcessBase):
	def __init__(self, wdir, rq):
		PluginProcessBase.__init__(self, wdir)
		if rq == None:
			rq = ['', '']
		self.cmd_str = rq[0]
		self.req = rq[1]

	def parse_result(self, text, sig):
		text = re.split(r'\r?\n', text)

		res = []
		if self.cmd_str == 'GREP':
			for line in text:
				if line == '':
					continue
				line = ['<unknown>'] + line.split(':', 2)
				res.append(line)
		else:
			assert(false)
			res.append(['', '', '', 'PluginProcess.parse_result: FAILED'])
			return res

		res = CtagsCache.CtagsThread(sig).apply_fix(self.cmd_str, res, ['<unknown>'])
		return res

if __name__ == '__main__':
	import sys

	def slot_result(sym, res):
		print('slot_result:    ', [str(sym), res])
		sys.exit(0)
	def slot_result_dbg(cmd, res, err_str):
		print('slot_result_dbg:', [str(cmd), str(res).strip().split('\n'), str(err_str)])
	def slot_rebuild():
		print('slot_rebuild')

	res = PluginProcess('.').run_query_process(['ls'], 'ls')
	#res = PluginProcess('/home/anil/prj/ss/lin').run_query_process(['cscope', '-q', '-k', '-L', '-d', '-0', 'vdso'], 'ls')
	print('res', res)



#!/usr/bin/env python3

# Copyright (c) 2010 Anil Kumar
# All rights reserved.
#
# License: BSD 

import os

from PyQt6.QtWidgets import *
from PyQt6.QtGui import *
from PyQt6.QtCore import *

dir_prefix = None

class DirTab(QWidget):
	sig_show_file = pyqtSignal(str)

	def __init__(self, parent=None):
		QWidget.__init__(self)

		self.is_ft = False
		self.d = None

		# Tree view
		self.tmodel = QFileSystemModel()
		self.tmodel.directoryLoaded.connect(self.tmodel_dir_loaded)
		self.tmodel.setRootPath(QDir.rootPath())
		self.tview = QTreeView()
		self.tview.setHeaderHidden(True)
		# For proper horizaontal scroll bar
		self.tview.setTextElideMode(Qt.TextElideMode.ElideNone)
		self.tview.expanded.connect(self.tview_expanded_or_collapsed)
		self.tview.collapsed.connect(self.tview_expanded_or_collapsed)

		self.ted = QLineEdit()
		self.completer = QCompleter()
		self.completer.setCompletionMode(QCompleter.CompletionMode.PopupCompletion)
		self.ted.setToolTip("Current folder")
		self.ted.setCompleter(self.completer)
		self.tdbtn = QPushButton()
		self.tdbtn.setIcon(QApplication.style().standardIcon(QStyle.StandardPixmap.SP_DirIcon))
		self.tdbtn.setToolTip("Open folder for browsing")
		self.trbtn = QPushButton()
		self.trbtn.setIcon(QApplication.style().standardIcon(QStyle.StandardPixmap.SP_BrowserReload))
		self.trbtn.setToolTip("Reset to common top-level folder of file in list")

		self.completer.setModel(self.tmodel)
		self.tview.setModel(self.tmodel)

		thlay = QHBoxLayout()
		thlay.addWidget(self.ted)
		thlay.addWidget(self.tdbtn)
		thlay.addWidget(self.trbtn)
		tvlay = QVBoxLayout()
		tvlay.addLayout(thlay)
		tvlay.addWidget(self.tview)
		
		self.setLayout(tvlay)

		self.tview.activated.connect(self.tview_itemActivated)
		self.tdbtn.clicked.connect(self.change_btn_cb)
		self.ted.editingFinished.connect(self.ted_editingFinished)
		self.trbtn.clicked.connect(self.reset_btn_cb)

		self.hide_view_columns(self.tview)

	def tmodel_dir_loaded(self, path):
		self.tview.resizeColumnToContents(0)

	def tview_expanded_or_collapsed(self, inx):
		self.tview.resizeColumnToContents(0)

	def dir_name(self):
		return self.d

	def hide_view_columns(self, view):
		header = view.header()
		for col in range(header.count()):
			if col > 0:
				view.setColumnHidden(col, True)


	def tview_itemActivated(self):
		list = self.tview.selectionModel().selectedIndexes()
		for item in list:
			if self.tmodel.fileInfo(item).isFile():
				self.sig_show_file.emit(self.tmodel.fileInfo(item).absoluteFilePath())
	
	def set_tab_name(self, dirstr):
		# Set tab name
		inx = self.parent.indexOf(self)
		try:
			if dirstr != dir_prefix:
				name = os.path.split(dirstr)[1]
			else:
				name = ''
			self.parent.setTabText(inx, name)
		except:
			pass

	def dir_reset(self, dirstr):
		self.set_tab_name(dirstr)
		self.d = dirstr
		self.tview.setRootIndex(self.tmodel.index(dirstr))
		self.ted.setText(dirstr)
		
	def ted_editingFinished(self):
		path = str(self.ted.text())
		if not os.path.isdir(path):
			path = os.path.dirname(path)
		if os.path.isdir(path):
			self.dir_reset(path)
		else:
			self.dir_reset(dir_prefix)

	def change_btn_cb(self):
		fdlg = QFileDialog(None, "Choose directory to browse")
		fdlg.setFileMode(QFileDialog.FileMode.Directory)
		fdlg.setOptions(QFileDialog.ShowDirsOnly) # and QFileDialog.HideNameFilterDetails)
		fdlg.setDirectory(self.ted.text())
		if (fdlg.exec()):
			browse_dir = fdlg.selectedFiles()[0]
			self.dir_reset(str(browse_dir))

	def reset_btn_cb(self):
		self.dir_reset(dir_prefix)

	def resizeEvent(self, event):
		self.tdbtn.setMaximumHeight(self.ted.height())
		self.trbtn.setMaximumHeight(self.ted.height())

class FileTab(QWidget):
	sig_show_file = pyqtSignal(str)
	
	def __init__(self, parent=None):
		QWidget.__init__(self)

		self.is_ft = True

		# List view
		self.le = QLineEdit()

		self.lview = QTreeWidget(self)
		self.lview.setColumnCount(2)
		self.lview.setHeaderLabels(['File', 'Path'])
		self.lview.setFont(QFont("San Serif", 8))
		self.lview.setIndentation(-2)
		self.lview.setAllColumnsShowFocus(True)

		lvlay = QVBoxLayout()
		lvlay.addWidget(self.le)
		lvlay.addWidget(self.lview)
		self.setLayout(lvlay)

		self.le.textChanged.connect(self.le_textChanged)
		self.le.returnPressed.connect(self.le_returnPressed)
		self.lview.itemActivated.connect(self.lview_itemActivated)

		global dir_prefix
		dir_prefix = QDir.rootPath() 

	def le_textChanged(self, text):
		if (text == ''):
			return
		self.lview.keyboardSearch('')
		self.lview.keyboardSearch(text)

	def le_returnPressed(self):
		self.le.clear()
		items = self.lview.selectedItems()
		if len(items) == 0:
			return
		self.lview.itemActivated.emit(items[0], 0)

	def lview_itemActivated(self, item):
		filename = str(item.data(1, Qt.ItemDataRole.DisplayRole))
		if self.is_rel_path:
			filename = filename.replace("...", dir_prefix, 1)
		self.sig_show_file.emit(filename)

	def keyPressEvent(self, ev):
		if ev.key() in [Qt.Key.Key_Up, Qt.Key.Key_Down, Qt.Key.Key_PageUp or Qt.Key.Key_PageDown]:
			self.lview.keyPressEvent(ev)
			return

	def search_file_cb(self):
		self.le.setFocus()
		self.le.selectAll()

	def clear(self):
		global dir_prefix
		dir_prefix = QDir.rootPath() 
		self.is_rel_path = False
		self.le.clear()
		self.lview.clear()

	def add_files(self, flist):
		global dir_prefix
		self.clear()
		dir_prefix = os.path.dirname(os.path.commonprefix(flist))
		if len(dir_prefix) > 16:
			self.is_rel_path = True
		for f in flist:
			if self.is_rel_path:
				f =  f.replace(dir_prefix, "...", 1)
			item = QTreeWidgetItem([os.path.basename(f), f])
			self.lview.addTopLevelItem(item)
			#if (self.lview.topLevelItemCount() > 0):
				#self.lview.resizeColumnToContents(0)
				#self.lview.resizeColumnToContents(1)
		self.lview.sortByColumn(0, Qt.SortOrder.AscendingOrder)
		self.lview.resizeColumnToContents(0)

class FileTree(QTabWidget):
	sig_show_file = pyqtSignal(str)
	
	def __init__(self, parent=None):
		QTabWidget.__init__(self)
		
		self.setMovable(True)

		t = FileTab()
		icon = QApplication.style().standardIcon(QStyle.StandardPixmap.SP_FileDialogDetailedView)
		self.addTab(t, icon, '')
		self.ft = t

		self.dlist = []

		self.new_dir_tab_cb()

		self.clear()

		# setup popup menu
		self.pmenu = QMenu()
		self.pmenu.addAction("&New Dir View", self.new_dir_tab_cb)
		self.pmenu.addAction("&Close Active Dir View", self.close_active_dir_tab_cb)
		self.pmenu.addAction("&Close All Dir View", self.close_all_dir_tab_cb)

	def set_tab_dir(self, t, d):
		if d == None or d == dir_prefix:
			t.reset_btn_cb()
		else:
			t.dir_reset(d)

	def new_dir_tab_cb(self, d=None):
		t = DirTab()
		t.parent = self
		self.dlist.append(t)
		icon = QApplication.style().standardIcon(QStyle.StandardPixmap.SP_DirClosedIcon)
		self.addTab(t, icon, '')
		self.set_tab_dir(t, d)

	def close_all_dir_tab_cb(self):
		for t in self.dlist:
			inx = self.indexOf(t)
			self.removeTab(inx)
		self.dlist = []
		# Always have atleast one dir view
		self.new_dir_tab_cb()

	def get_dir_view_list(self):
		dv_list = []
		for t in self.dlist:
			d = t.dir_name()
			if d != dir_prefix:
				dv_list.append(d)
		return dv_list

	def open_dir_view(self, filename):
		d = filename
		if not os.path.isdir(filename):
			d = os.path.dirname(str(filename))
		self.new_dir_tab_cb(d)

	def close_active_dir_tab_cb(self):
		inx = self.currentIndex()
		if inx < 0:
			return
		t = self.widget(inx)
		if t.is_ft:
			return
		if self.count() <= 2:
			self.close_all_dir_tab_cb()
			return
		self.dlist.remove(t)
		self.removeTab(inx)

	def addTab(self, t, icon, x):
		t.sig_show_file.connect(self.sig_show_file)
		QTabWidget.addTab(self, t, icon, x)
		if self.count() > 2:
			inx = self.indexOf(t)
		else:
			inx = 0
		self.setCurrentIndex(inx)

	def mousePressEvent(self, m_ev):
		QTabWidget.mousePressEvent(self, m_ev)
		if (m_ev.button() == Qt.MouseButton.RightButton):
			self.pmenu.exec(QCursor.pos())

	def search_file_cb(self):
		self.ft.search_file_cb()
		self.setCurrentWidget(self.ft)

	def clear(self):
		self.ft.clear()
		self.close_all_dir_tab_cb()

	def add_files(self, flist):
		old_dir_prefix = dir_prefix
		self.ft.add_files(flist)
		for t in self.dlist:
			d = t.dir_name()
			if d == old_dir_prefix:
				d = dir_prefix
			self.set_tab_dir(t, d)



# Copyright (c) 2010 Anil Kumar
# All rights reserved.
#
# License: BSD 

import sys
import os
import string

from PyQt6 import QtWidgets, QtGui, QtCore, uic

from PyQt6.QtWidgets import *
from PyQt6.QtGui import *
from PyQt6.QtCore import *

class DebugInfoEntry(QFrame):
	def __init__(self, parent=None):
		QFrame.__init__(self)
		self.vlay = QVBoxLayout()
		self.setLayout(self.vlay)

	#def add_result_continue(self):
		#self.add_result(self.name, self.res)

	def add_result(self, cmd, out, err):
		#self.setFrameStyle(QFrame.StyledPanel | QFrame.Plain)
		self.cmd_lbl = QLabel()
		self.cmd_lbl.setText(cmd)
		self.vlay.addWidget(self.cmd_lbl)

		#if err != None and err != '':
			#self.err_lbl = QLabel()
			#self.err_lbl.setText('Error: ' + err)
			#self.vlay.addWidget(self.err_lbl)

		#self.res_lbl = QLabel()
		#self.res_lbl.setText(out)
		#self.vlay.addWidget(self.res_lbl)
		#self.res_lbl.hide()

class DebugWindow(QMainWindow):
	dlg = None

	def __init__(self, parent=None):
		QMainWindow.__init__(self, parent)

		self.ui = uic.loadUi('ui/debug.ui', self)
		self.vlay = self.ui.dbg_widget.layout()

	def append_widget(self, w):
		self.vlay.addWidget(w)

	def run_dialog(self):
		self.ui.show()
		vbar = self.ui.dbg_scroll_area.verticalScrollBar()
		vbar.setValue(vbar.maximum())
		self.ui.show()
	
	def closeEvent(self, e):
		DebugWindow.dlg = None

def show_dbg_dialog(parent):
	if (DebugWindow.dlg == None):
		DebugWindow.dlg = DebugWindow(parent)
	DebugWindow.dlg.run_dialog()


def connect_to_sig(sig_res):
	if (DebugWindow.dlg == None):
		return
	entry = DebugInfoEntry()
	DebugWindow.dlg.append_widget(entry)
	sig_res.connect(entry.add_result)



# Copyright (c) 2010 Anil Kumar
# All rights reserved.
#
# License: BSD 

import os
import sys
import re

from PyQt6 import QtWidgets, QtGui, QtCore, uic

from PyQt6.QtWidgets import *
from PyQt6.QtGui import *
from PyQt6.QtCore import *

class ProjectSettingsDialog(QDialog):
	def __init__(self):
		QDialog.__init__(self)

		self.ui = uic.loadUi('ui/project_settings.ui', self)

	def run_dialog(self, prj_type, prj_args):
		if prj_args:
			prj_dir = prj_args[0]
			self.pi_type_lbl.setText(prj_type)
			self.pi_path_lbl.setText(prj_dir)
			self.exec()
			return None
		fdlg = QFileDialog(None, "Choose source code directory")
		fdlg.setFileMode(QFileDialog.FileMode.Directory);
		#fdlg.setDirectory(self.pd_path_inp.text())
		if fdlg.exec():
			d = fdlg.selectedFiles()[0];
			d = str(d)
			if not d:
				return None
			d = os.path.normpath(d)
			if d == '' or not os.path.isabs(d):
				return None
			proj_args = [d, None, None]
			return proj_args
		return None

def show_settings_ui(prj_type, proj_args):
	dlg = None
	if prj_type == 'cscope':
		from . import CscopeProjectUi
		dlg = CscopeProjectUi.CscopeProjectSettingsDialog()
	else:
		dlg = ProjectSettingsDialog()
	return dlg.run_dialog(prj_type, proj_args)




#!/usr/bin/env python

# Copyright (c) 2010 Anil Kumar
# All rights reserved.
#
# License: BSD 

import sys
import subprocess
import os
import re

class CtagsInhCache:
	def __init__(self, is_base, is_fq, is_debug):
		self.cit_cache = {}
		self.is_base   = is_base
		self.is_fq     = is_fq
		self.is_debug  = False

		self.ct_opt_I_file = os.getenv('SEASCOPE_CTAGS_OPT_I_FILE')
		if self.ct_opt_I_file:
			if not os.path.isfile(self.ct_opt_I_file):
				self.ct_opt_I_file = None

		map = os.getenv('SEASCOPE_CTAGS_SUFFIX_CMD_MAP', 0)
		if map:
			try:
				map = eval(map)
			except:
				print('SEASCOPE_CTAGS_SUFFIX_CMD_MAP has errors')
				map = None
		self.ct_custom_map = map


	def _filterCtInherits(self, data, sym=None):
		res = {}
		for line in data:
			if line == '':
				continue
			line = line.split('\t', 4)
			if len(line) == 4:
				continue
			_sd = dict([ x.split(':', 1) for x in line[4].split('\t')])
			if 'inherits' not in _sd:
				continue
			sd = _sd['inherits'].strip()
			if sd == '':
				continue
			if 'class' in _sd:
				cls = _sd['class'].strip()
			else:
				cls = None
			if line[1] not in res:
				res[line[1]] = []
			res[line[1]].append([line[0], sd, cls])
			if self.is_debug:
				print(line[0], line[1], sd, cls)
		return res

	def _runCtagsCustom(self, fl):
		if not self.ct_custom_map:
			return []
		cmd_list = []
		for (suffix, cmd) in self.ct_custom_map:
			_fl = [ f for f in fl if f.endswith(suffix) ]
			args = cmd.split()
			args += _fl
			cmd_list.append(args)

		if not len(cmd_list):
			return []

		out_data_all = []
		for args in cmd_list:
			proc = subprocess.Popen(args, stdin=subprocess.PIPE, stdout=subprocess.PIPE)
			(out_data, err_data) = proc.communicate('\n'.join(fl).encode())
			out_data = out_data.decode()
			out_data = re.split(r'\r?\n', out_data)
			out_data_all += out_data
		return out_data_all

	def _runCtags(self, fl):
		cmd = 'ctags -n -u --fields=+i -L - -f -'
		if self.ct_opt_I_file:
			cmd += ' -I ' + self.ct_opt_I_file
		args = cmd.split()
		proc = subprocess.Popen(args, stdin=subprocess.PIPE, stdout=subprocess.PIPE)
		(out_data, err_data) = proc.communicate('\n'.join(fl).encode())
		out_data = out_data.decode()
		out_data = re.split(r'\r?\n', out_data)
		out_data += self._runCtagsCustom(fl)
		return out_data

	def runCtagsInh(self, fl):
		data = self._runCtags(fl)
		ct_dict = self._filterCtInherits(data, None)
		for f in fl:
			if f not in ct_dict:
				ct_dict[f] = []
		return ct_dict

	def ctInhInfo(self, flist):
		fl = [ f for f in flist if f not in self.cit_cache ]
		#print len(fl), len(flist), len(self.cit_cache)
		if len(fl):
			ct_dict = self.runCtagsInh(fl)
			self.cit_cache.update(ct_dict)
		res = []
		for f in flist:
			res += self.cit_cache[f]
		return res

class ClassGraphGenerator:
	def __init__(self, d, pcmd=None, wlimit=5000, is_base=False):
		self.wdir = d;
		self.pcmd = pcmd
		self.width_limit = wlimit;
		self.dname = None
		self.dirCtInhInfo = None
		self.graphRules = []
		self.visitedRules = {}
		self.visitedSym = {}
		if is_base:
			self.visitedSym = {'object' : 1}
			if os.getenv('SEASCOPE_CTAGS_SUFFIX_CMD_MAP', 0):
				for c in [ 'Entity', 'Type', 'PtrInterFace','Enum','Nominal' ]:
					self.visitedSym[c] = 1
					self.visitedSym['Tac::' + c] = 1
		self.is_base = is_base
		self.is_fq = False
		self.is_debug = os.getenv('SEASCOPE_CLASS_GRAPH_DEBUG', 0)
		self.cic = CtagsInhCache(self.is_base, self.is_fq, self.is_debug)

	def addGraphRule(self, sym, d):
		if (sym, d) in self.visitedRules:
			return
		self.visitedRules[(sym, d)] = True
		self.graphRules.append([sym, d])

	def refFiles(self, sym):
		args = list(self.pcmd) + [sym]
		try:
			# In python >= 2.7 can use subprocess.check_output
			# output = subprocess.check_output(args, cwd=self.wdir)
			proc = subprocess.Popen(args, cwd=self.wdir, stdout=subprocess.PIPE)
			(output, err_data) = proc.communicate()
			output = output.decode()
			output = re.split(r'\r?\n', output)
		except Exception as e:
			print('dir:', self.wdir, ':cmd:', args, ':', e, '\n', file=sys.stderr)
			sys.exit(-1)
		res = set()
		for line in output:
			if line == '':
				continue
			f = line.split(':', 1)[0]
			f = os.path.normpath(os.path.join(self.wdir, f))
			res.add(f)
		return res

	def parseCtagsInherits(self, data, sym=None):
		res = []
		for line in data:
			sd = line[1]
			if sd == '':
				continue
			sd = sd.split(',')
			if self.is_fq:
				dd = [ x.strip() for x in sd ]
				cls = line[2]
			else:
				dd = [ re.split(r'::|\.', x.strip())[-1] for x in sd ]
				cls = None
			if not sym:
				res.append([line[0], dd])
				continue
			if self.is_base:
				if cls:
					if sym == cls + "::" + line[0]:
						res += dd
					if sym == cls + "." + line[0]:
						res += dd
				else:
					if sym == line[0]:
						res += dd
			else:
				if sym in dd:
					if cls:
						sep = '::'
						if '.' in cls:
							sep = '.'
						res.append(cls + sep + line[0])
					else:
						res.append(line[0])
		if self.is_debug:
			if sym:
				print(sym, res)
		return res


	def classHierarchy(self, sym):
		if self.is_fq:
			subSym = re.split(r'::|\.', sym)[-1]
		else:
			subSym = sym

		if self.pcmd:
			fl = self.refFiles(subSym)
			data = self.cic.ctInhInfo(fl)
		else:
			data = self.dirCtInhInfo

		res = self.parseCtagsInherits(data, sym)
		return res

	def runCtInhForDir(self, dname):
		self.dname = dname
		fl = []
		for root, dirs, files in os.walk(dname, followlinks=True):
			fl += [os.path.join(root, f) for f in files]
		self.dirCtInhInfo = self.cic.ctInhInfo(fl)

	def classHierarchyRecursive(self, symList):
		for sym in symList:
			if sym in self.visitedSym:
				continue
			self.visitedSym[sym] = 1

			dclasses = self.classHierarchy(sym)
			if len(dclasses):
				if len(dclasses) > self.width_limit:
					if is_base:
						s = 'base'
					else:
						s= 'derived'
					print('num %s classes(%s) = %d, truncating to %d.\n' % (s, sym, len(dclasses), self.width_limit), file=sys.stderr)
					dclasses = dclasses[0:self.width_limit]
					self.addGraphRule(sym, '...(%s)' % sym)
				for d in dclasses:
					self.addGraphRule(sym, d)
				self.classHierarchyRecursive(dclasses)

	def classHierarchyForDir(self, dname):
		res = self.parseCtagsInherits(self.dirCtInhInfo)
		for (d, blist) in res:
			for b in blist:
				self.addGraphRule(b, d)

	def prepareDotInput(self, sym_or_dname):
		if len(self.graphRules) == 0:
			if is_base:
				s = 'base'
			else:
				s= 'derived'
			print('No %s classes for %s\n' % (s, sym_or_dname), file=sys.stderr)
			sys.exit(0)
		dotInput = 'digraph "%s" {\n' % sym_or_dname
		if not sym_or_dname:
			dotInput += '\t"%s" [style=bold];\n' % sym_or_dname
		for r in self.graphRules:
			if not self.is_base:
				dotInput += '\t"%s" -> "%s";\n' % (r[0], r[1])
			else:
				dotInput += '\t"%s" -> "%s";\n' % (r[1], r[0])
		dotInput += '}\n'
		return dotInput

	def saveDotFile(self, sym, dotInput):
		f = open(sym + '.dot', 'w')
		f.write(dotInput)
		f.close()

	def saveSvgFile(sym, svg_data):
		dot_svg = sym + '.svg'
		f = open(dot_svg, 'w')
		f.write(svg_data)
		f.close()

		print('saved', dot_svg, '\n', file=sys.stderr)

	def generateGraph(self, dname, sym=None):
		if not self.pcmd:
			self.runCtInhForDir(dname)
		if sym:
			if sym == '::' or sym == '.':
				return
			if re.search(r'::|\.', sym):
				self.is_fq = True
			if sym.startswith('::') or sym.startswith('.'):
				sym = re.split(r'::|\.', sym, maxsplit=1)[-1]
			self.classHierarchyRecursive([sym])
			dotInput = self.prepareDotInput(sym)
		else:
			self.classHierarchyForDir(dname)
			dotInput = self.prepareDotInput(dname)

		if self.is_debug:
			print(dotInput)
			return ''
		#self.saveDotFile(sym, dotInput)

		args = ['dot', '-Tsvg']
		try:
			p = subprocess.Popen(args, stdin=subprocess.PIPE, stdout=subprocess.PIPE)
			(svg_data, err_data) = p.communicate(dotInput.encode())
			svg_data = svg_data.decode()
			if err_data and err_data.decode() != '':
				print(err_data, '\n', file=sys.stderr)
			#self.saveSvgFile(sym, svg_data)
			return svg_data
		except Exception as e:
			print('Failed to run:', ' '.join(args), '\n', file=sys.stderr)
			print(e, '\n', file=sys.stderr)


if __name__ == '__main__':
	import optparse
	usage = "usage: %prog [options] (-d <code_dir/file> | -t <prj_type>) [symbol]"
	op = optparse.OptionParser(usage=usage)
	op.add_option("-b", action="store_true", dest="is_base", help="Show base classes")
	op.add_option("-d", "--codedir", dest="code_dir", help="Code dir", metavar="CODE_DIR")
	op.add_option("-t", "--type", dest="prj_type", help="project type: idutils|gtags|cscope", metavar="PRJ_TYPE")
	(options, args) = op.parse_args()

	# dname
	if not options.code_dir:
		print('Specify -d', file=sys.stderr)
		sys.exit(-1)
	dname = options.code_dir
	if not os.path.exists(dname):
		print('"%s": does not exist' %  dname, file=sys.stderr)
		sys.exit(-2)
	wdir = dname
	if not os.path.isdir(wdir):
		wdir = os.path.dirname(wdir)

	# sym
	sym = None
	if len(args):
		if len(args) != 1:
			print('Please specify only one symbol', file=sys.stderr)
			sys.exit(-3)
		sym = args[0]

	# ptype
	ptype = options.prj_type
	if ptype:
		if not sym:
			print('-t option needs sepficfying symbol', file=sys.stderr)
			sys.exit(-4)
		if not os.path.isdir(dname):
			print('-t option needs codedir to be a directory', file=sys.stderr)
			sys.exit(-5)
	pcmd = None
	if ptype and os.path.isdir(dname):
		prj_list = [
			['idutils', 'ID',        ['lid', '-R', 'grep', '--']                          ],
			['gtags', 'GRTAGS',      ['global', '-a', '--result=grep', '-x', '-r', '--']  ],
			['cscope', 'cscope.out', ['cscope', '-L', '-d',  '-0', '--']                  ],
			['grep',   '',           ['grep', '-R', '-n', '-I', '--',]                    ],
		]
		for p in prj_list:
			if p[0] == ptype:
				if os.path.exists(os.path.join(dname, p[1])):
					pcmd = p[2]
				break
	# is_base
	is_base = options.is_base

	# run
	cgg = ClassGraphGenerator(wdir, pcmd=pcmd, is_base=is_base)
	svg_data = cgg.generateGraph(dname, sym)

	print(svg_data)



#!/usr/bin/env python

# Copyright (c) 2012 Anthony Liu
# All rights reserved.
#
# License: BSD 

import sys
from array import *

class CodemarkManager():

	def __init__(self, parent=None):
		self.haystack = []
		
	def check(self, filename, line):
		if self.haystack.count((filename, line)):
			return 1
		return 0

	def append(self, filename, line):
		if self.check(filename, line) == 0:
			self.haystack.append((filename, line))
		return self.haystack.index((filename, line))
	
	def delete(self, filename, line):
		if self.haystack.count((filename, line)) == 0:
			return -1
		else:
			self.haystack.remove((filename, line))
			return 0
	
	def delete_index(self, index):
		if self.count() <= index:
			return -1
		elif index < 0:
			return -1
		else:
			self.haystack.pop(index)
			return 0
	
	def count(self):
		return len(self.haystack)
	
	def get(self, index):
		if self.count() <= index:
			return (None, None)
		elif index < 0:
			return (None, None)
		else:
			(filename, line) = self.haystack.pop(index)
			self.haystack.insert(index, (filename, line))
			return (filename, line)
		
	def clear(self):
		del self.haystack[:]

	def codemarks(self):
		return self.haystack

	def dump(self):
		for i in range(len(self.haystack)):
			print("cm index ", i, " ", self.haystack[i], file=sys.stderr)

#############################################################################
#
# Basic unit test
#
#############################################################################

if __name__ == '__main__':
	
	cm = CodemarkManager()

	#
	# case 1, basic add, delete
	#
	
	print("Initial empty", file=sys.stderr)
	
	print("cm has ", cm.count(), "items (0)", file=sys.stderr)

	cm.append("file1", 100)
	cm.append("file2", 200)
	cm.append("file3", 300)

	print("cm has " , cm.count(), "items (3) ", file=sys.stderr)
	
	cm.delete("file1", 100)

	print("cm has " , cm.count(), "items (2)", file=sys.stderr)
	
	cm.delete("file2", 100) # not existed

	print("cm has " , cm.count(), "items (2)", file=sys.stderr)

	cm.delete_index(1)

	print("cm has " , cm.count(), "items (1)", file=sys.stderr)

	#
	# case 2, dump
	#

	for i in range(cm.count()):
		(f, l) = cm.get(i)
		print(f, " (#", l, ")", file=sys.stderr)

	#
	# case 3, duplicate
	#

	index = cm.append("file4", 400)
	print("index for file4 #400 is " , index, file=sys.stderr)
	
	index = cm.append("file4", 400)
	print("index for file4 #400 is " , index, " (again)", file=sys.stderr)
	
	print("cm has " , cm.count(), "items (2) ", file=sys.stderr)

	#
	# case 4, dump for eye examine
	#

	print("cm dump for final check", file=sys.stderr)

	cm.dump()



#!/usr/bin/env python3

# Copyright (c) 2010 Anil Kumar
# All rights reserved.
#
# License: BSD 

from PyQt6.QtWidgets import *
from PyQt6.QtGui import *
from PyQt6.QtCore import *

class CallTreeWidgetItem(QTreeWidgetItem):
	def __init__(self, li):
		QTreeWidgetItem.__init__(self, li)
		func = li[0]
		if (func == "<global>"):
			self.setChildIndicatorPolicy(QTreeWidgetItem.ChildIndicatorPolicy.DontShowIndicatorWhenChildless)
		else:
			self.setChildIndicatorPolicy(QTreeWidgetItem.ChildIndicatorPolicy.ShowIndicator)
		self.is_done = False

	def set_no_children(self):
		self.setChildIndicatorPolicy(QTreeWidgetItem.ChildIndicatorPolicy.DontShowIndicatorWhenChildless)

	def column_val(self, col):
		return str(self.data(col, Qt.ItemDataRole.DisplayRole))

	def add_result(self, res):
		self.is_done = True
		res_list = []
		ret_val = True
		for line in res:
			if (line[0] == "<global>"):
				continue
			if (line[0] == self.column_val(0)):
				if (line [1] == self.column_val(1) and line[2] == self.column_val(2)):
					continue
			self.addChild(CallTreeWidgetItem(line))
		if (self.childCount() == 0):
			self.set_no_children()
		# resize column
		self.treeWidget().resizeColumnToContents(0)
		self.treeWidget().resizeColumnToContents(1)
		self.treeWidget().resizeColumnToContents(2)
		self.treeWidget().resizeColumnToContents(3)
		# hdr
		#self.treeWidget().header().setDefaultAlignment(Qt.AlignmentFlag.AlignRight)
		return ret_val

class CallTreeWidget(QTreeWidget):
	sig_show_file_line = pyqtSignal(str, int)

	def __init__(self, parent, cmd_func, cmd_id, cmd_opt, hint_file):
		QTreeWidget.__init__(self, parent)
		self.is_busy = False
		self.cmd_func = cmd_func
		self.cmd_id = cmd_id
		self.cmd_opt = cmd_opt
		self.hint_file = hint_file
		
		self.itemExpanded.connect(self.ctree_itemExpanded)
		self.itemActivated.connect(self.ctree_itemActivated)

		self.setExpandsOnDoubleClick(False)
		self.setColumnCount(4)
		self.setHeaderLabels(['Tag', 'File', 'Line', 'Text'])

		self.setSelectionMode(QAbstractItemView.SelectionMode.SingleSelection)
		#self.resize(QSize(800, 500))
		self.setMinimumWidth(800)
		self.setMinimumHeight(500)

		##sel behaviour
		#self.setSelectionBehavior(QAbstractItemView.SelectRows)
		## set the font
		self.setFont(QFont("San Serif", 8))

		#self.setTextElideMode(Qt.TextElideMode.ElideLeft)
		self.setAllColumnsShowFocus(True)

	def add_root(self, name):
		parent = CallTreeWidgetItem([name, '', '', ''])
		self.addTopLevelItem(parent)
		parent.setExpanded(True)

	def mousePressEvent(self, m_ev):
		QTreeWidget.mousePressEvent(self, m_ev)
		if (m_ev.button() == Qt.MouseButton.RightButton):
			self.last_minx = self.indexAt(m_ev.pos())

	def ctree_itemActivated(self, item, col):
		filename = item.column_val(1)
		try:
			line = int(item.column_val(2))
		except:
			return
		self.sig_show_file_line.emit(filename, line)

	def ctree_itemExpanded(self, item):
		if (item.is_done):
			return
		tag = str(item.data(0, Qt.ItemDataRole.DisplayRole))
		if str(item.data(1, Qt.ItemDataRole.DisplayRole)) == '':
			opt = self.cmd_opt
		else:
			opt = None
		if (self.is_busy):
			return
		self.is_busy = True
		self.pbar = QProgressBar(self)
		self.pbar.setMinimum(0)
		self.pbar.setMaximum(0)
		self.pbar.show()
		
		## add result
		rquery = {}
		rquery['cmd'] = self.cmd_id
		rquery['req'] = tag
		rquery['opt'] = opt
		hfile = str(item.data(1, Qt.ItemDataRole.DisplayRole))
		if hfile == '':
			hfile = self.hint_file
		rquery['hint_file'] = hfile
		sig_res = self.cmd_func(rquery)
		self.query_item = item
		sig_res[0].connect(self.ctree_add_result)

	def ctree_add_result(self, req, res):
		self.query_item.add_result(res)
		self.is_busy = False
		if self.pbar:
			self.pbar.setParent(None)
			self.pbar = None


class CallTreeWindow(QMainWindow):
	sig_show_file_line = pyqtSignal(str, int)
	parent = None

	def __init__(self, req, cmd_func, cmd_args, cmd_opt, hint_file):
		QMainWindow.__init__(self, CallTreeWindow.parent)
		self.req = req

		self.setWindowTitle(req)

		self.setFont(QFont("San Serif", 8))

		w = QWidget()
		self.setCentralWidget(w)
		self.vlay = QVBoxLayout()
		w.setLayout(self.vlay)

		self.sw = QStackedWidget()

		self.hlay = QHBoxLayout()
		self.vlay.addLayout(self.hlay)
		self.vlay.addWidget(self.sw)
		
		self.bgrp = QButtonGroup()
		self.bgrp.buttonClicked.connect(self.set_current)
		self.bgrp.setExclusive(True)

		self.btn = []
		self.ctree = []
		for inx in range(len(cmd_args)):
			# cmd format: [ cmd_id, cmd_str, cmd_tip ]
			cmd = cmd_args[inx]

			btn = QToolButton()
			btn.setText(cmd[1])
			btn.setToolTip(cmd[2])
			#btn.setFlat(True)
			btn.setCheckable(True)
			self.bgrp.addButton(btn, inx)
			self.hlay.addWidget(btn)

			ct = CallTreeWidget(self, cmd_func, cmd[0], cmd_opt, hint_file)
			ct.sig_show_file_line.connect(self.sig_show_file_line)
			self.sw.addWidget(ct)

			self.btn.append(btn)
			self.ctree.append(ct)
		self.hlay.addStretch(0)
		self.set_current(self.btn[0])

	def set_current(self, btn):
		inx = self.bgrp.id(btn)
		self.btn[inx].setChecked(True)
		self.sw.setCurrentIndex(inx)
		ct = self.ctree[inx]
		if ct.topLevelItemCount() == 0:
			ct.add_root(self.req)
		ct.setFocus()

def ctree_show_file_line(filename, line):
	parent = CallTreeWindow.parent
	parent.raise_()
	parent.activateWindow()
	parent.show_file_line(filename, line)

def create_page(req, cmd_func, cmd_args, cmd_opt, hint_file):
	w = CallTreeWindow(req, cmd_func, cmd_args, cmd_opt, hint_file)
	w.sig_show_file_line.connect(ctree_show_file_line)
	w.show()



#!/usr/bin/env python3

# Copyright (c) 2010 Anil Kumar
# All rights reserved.
#
# License: BSD 

from PyQt6.QtWidgets import *
from PyQt6.QtGui import *
from PyQt6.QtCore import *
from PyQt6.QtSvgWidgets import *
import os
import sys

if __name__ == '__main__':
	import sys
	import os
	app_dir = os.path.dirname(os.path.realpath(__file__))
	os.chdir(app_dir)
	sys.path.insert(0, os.path.abspath('..'))
	os.chdir(os.path.abspath('..'))

from backend.plugins.PluginBase import PluginProcess

class ClassGraphProcess(PluginProcess):
	def __init__(self, wdir, rq):
		PluginProcess.__init__(self, wdir, rq)
		self.name = 'call graph process'

	def parse_result(self, text, sig):
		return [text]


class ClassGraphWidget(QWidget):
	def __init__(self, parent, cmd_func, cmd_id, cmd_opt):
		QWidget.__init__(self, parent)
		self.is_busy = False
		self.is_done = False
		self.cmd_func = cmd_func
		self.cmd_id = cmd_id
		self.cmd_opt = cmd_opt
		self.is_debug = os.getenv('SEASCOPE_CLASS_GRAPH_VIEW_DEBUG')

		self.vlay1 = QVBoxLayout()
		self.setLayout(self.vlay1)
		#self.hlay1 = QHBoxLayout()
		#self.vlay1.addLayout(self.hlay1)
		#self.add_buttons(self.hlay1)
		self.lbl = QLabel()
		self.vlay1.addWidget(self.lbl)

		self.vlay2 = QVBoxLayout()
		self.scrolla = QScrollArea()
		self.scrolla.setLayout(self.vlay2)
		self.vlay1.addWidget(self.scrolla)

	def startQuery(self, req, dname, proj_type, inx):
		if self.is_done:
			return

		name = req if req else dname
		self.lbl.setText(['derived', 'base'][inx] + '(' + name + ')')

		tool_path = os.path.join('tools', 'ClassGraph.py')
		pargs = [sys.executable, tool_path]
		if inx == 1:
			pargs += ['-b']
		pargs += ['-d', dname]
		if proj_type:
			pargs += ['-t', proj_type]
		if req:
			pargs += [req]
		sig_res = ClassGraphProcess('.', None).run_query_process(pargs, req)
		sig_res[0].connect(self.clgraph_add_result)
		self.is_busy = True
		self.show_progress_bar()

	def set_current(self, btn):
		inx = self.bgrp.id(btn)
		#self.btn[inx].setChecked(True)
		#print 'inx clicked', inx
		if inx == 0:
			print(self.svgw.renderer().defaultSize())
			self.svgw.setMinimumSize(0, 0)
			self.svgw.setMinimumSize(self.svgw.sizeHint())
			#self.svgw.setMaximumSize(self.svgw.sizeHint())
			print(self.scrolla.sizeHint())
		if inx == 1:
			print(self.svgw.renderer().defaultSize())
			self.svgw.setMinimumSize(0, 0)
			#self.svgw.setMaximumSize(self.svgw.sizeHint())
			print(self.scrolla.sizeHint())
		if inx == 2:
			print(self.svgw.renderer().defaultSize())
			self.svgw.setMinimumSize(0, 0)
			self.svgw.resize(self.scrolla.size())
			#self.svgw.setMaximumSize(self.svgw.sizeHint())
			print(self.scrolla.sizeHint())

	def add_buttons(self, hlay):
		self.bgrp = QButtonGroup()
		self.bgrp.buttonClicked.connect(self.set_current)
		self.bgrp.setExclusive(True)
		for inx in range(3):
			btn = QToolButton()
			btn.setText(str(inx))
			btn.setToolTip(str(inx))
			#btn.setFlat(True)
			btn.setCheckable(True)
			self.bgrp.addButton(btn, inx)
			hlay.addWidget(btn)

	def clgraph_add_result(self, req, res):
		if self.is_debug:
			print(res)
		self.is_busy = False
		self.is_done = True
		self.remove_progress_bar()
		
		self.svgw = QSvgWidget()
		self.scrolla.setWidget(self.svgw)
		self.svgw.load(QByteArray(res[0].encode()))
		
		#print self.svgw.renderer().defaultSize()
		sz = self.svgw.sizeHint()
		scale = 1
		if sz.width() > 1024:
			scale = 0.8
		min_w = int(sz.width() * scale)
		min_h = int(sz.height() * scale)
		self.svgw.setMinimumSize(min_w, min_h)
		#self.svgw.setMaximumSize(self.svgw.sizeHint())
		#print self.scrolla.sizeHint()

	def show_progress_bar(self):
		self.pbar = QProgressBar(self.scrolla)
		self.pbar.setMinimum(0)
		self.pbar.setMaximum(0)
		self.pbar.show()

	def remove_progress_bar(self):
		if self.pbar:
			self.pbar.hide()
			self.pbar.setParent(None)
			self.pbar = None

class ClassGraphWindow(QMainWindow):
	parent = None

	def __init__(self, req, dname, proj_type, cmd_func, cmd_args, cmd_opt):
		QMainWindow.__init__(self, ClassGraphWindow.parent)
		self.req = req
		self.dname = dname
		self.proj_type = proj_type

		if req:
			self.setWindowTitle(req)
		else:
			self.setWindowTitle(dname)

		self.setFont(QFont("San Serif", 8))

		w = QWidget()
		self.setCentralWidget(w)
		self.vlay = QVBoxLayout()
		w.setLayout(self.vlay)

		self.sw = QStackedWidget()

		self.hlay = QHBoxLayout()
		self.vlay.addLayout(self.hlay)
		self.vlay.addWidget(self.sw)
		
		self.bgrp = QButtonGroup()
		self.bgrp.buttonClicked.connect(self.set_current)
		self.bgrp.setExclusive(True)

		self.btn = []
		self.ctree = []
		for inx in range(len(cmd_args)):
			# cmd format: [ cmd_id, cmd_str, cmd_tip ]
			cmd = cmd_args[inx]

			btn = QToolButton()
			btn.setText(cmd[1])
			btn.setToolTip(cmd[2])
			#btn.setFlat(True)
			btn.setCheckable(True)
			self.bgrp.addButton(btn, inx)
			self.hlay.addWidget(btn)

			ct = ClassGraphWidget(self, cmd_func, cmd[0], cmd_opt)
			self.sw.addWidget(ct)

			self.btn.append(btn)
			self.ctree.append(ct)
		self.hlay.addStretch(0)
		self.set_current(self.btn[0])

	def set_current(self, btn):
		inx = self.bgrp.id(btn)
		self.btn[inx].setChecked(True)
		self.sw.setCurrentIndex(inx)
		ct = self.ctree[inx]
		ct.setFocus()
		
		ct.startQuery(self.req, self.dname, self.proj_type, inx)



def create_page(req, dname, proj_type, cmd_func, cmd_args, cmd_opt):
	w = ClassGraphWindow(req, dname, proj_type, cmd_func, cmd_args, cmd_opt)
	w.resize(900, 600)
	w.show()
	return w

if __name__ == '__main__':
	import optparse
	usage = "usage: %prog (-d <code_dir/file> | -t <prj_type>) [symbol]"
	op = optparse.OptionParser(usage=usage)
	op.add_option("-d", "--codedir", dest="code_dir", help="Code dir", metavar="CODE_DIR")
	op.add_option("-t", "--type", dest="prj_type", help="project type: idutils|gtags|cscope", metavar="PRJ_TYPE")
	(options, args) = op.parse_args()

	# dname
	if not options.code_dir:
		print('Specify -d', file=sys.stderr)
		sys.exit(-1)
	dname = options.code_dir
	if not os.path.exists(dname):
		print('"%s": does not exist' %  dname, file=sys.stderr)
		sys.exit(-2)
	wdir = dname
	if not os.path.isdir(wdir):
		wdir = os.path.dirname(wdir)

	# sym
	sym = None
	if len(args):
		if len(args) != 1:
			print('Please specify only one symbol', file=sys.stderr)
			sys.exit(-3)
		sym = args[0]

	# ptype
	ptype = options.prj_type
	if ptype:
		if not sym:
			print('-t option needs sepficfying symbol', file=sys.stderr)
			sys.exit(-4)
		if not os.path.isdir(dname):
			print('-t option needs codedir to be a directory', file=sys.stderr)
			sys.exit(-5)
	pcmd = None
	if ptype and os.path.isdir(dname):
		prj_list = [
			['idutils', 'ID'         ],
			['gtags',   'GRTAGS'     ],
			['cscope',  'cscope.out' ],
			['grep',    ''           ],
		]
		for p in prj_list:
			if p[0] == ptype:
				if not os.path.exists(os.path.join(dname, p[1])):
					print('Failed to find "%s" in directory "%s"' % (p[1], dname), file=sys.stderr)
					sys.exit(-6)
					
	app = QApplication(sys.argv)
	cmd_args = [
		['CLGRAPH', 'D', 'Derived classes'],
		['CLGRAPH', 'B', 'Base classes']
	]

	w = create_page(sym, dname, ptype, None, cmd_args, None)

	sys.exit(app.exec())



#!/usr/bin/env python

# Copyright (c) 2013 dangbinghoo <dangbinghoo@gmail.com>
# All rights reserved.
#
# License: BSD

import os
import re
import array

from PyQt6 import QtWidgets, QtGui
from PyQt6.QtWidgets import *
from PyQt6.QtGui import *
from PyQt6.QtCore import *

from .EdView import EditorView

# code context view
class ContextEditorView(EditorView):
	sig_dblclick = pyqtSignal()

	def __init__(self, parent = None):
		EditorView.__init__(self, parent)

		self.setCaretLineBackgroundColor(QtGui.QColor("#ffaa7f"))
		
	def mouseReleaseEvent(self, ev):
		super(EditorView, self).mouseReleaseEvent(ev)
		
	def mousePressEvent(self, ev):
		pass
	
	def keyPressEvent(self, ev):
		pass
	
	def contextMenuEvent(self, ev):
		pass
	
	def mouseDoubleClickEvent(self, ev):
		self.sig_dblclick.emit()

	def show_file_line(self, fname, line):
		self.open_file(fname)
		self.goto_line(int(line))
		self.show()

class codeResultView(QTextBrowser):
	sig_result_view_openfile = pyqtSignal(str, int)
	def __init__(self, parent = None):
		QTextBrowser.__init__(self)
		self.setReadOnly(True)
		
	def showResultList(self, sym, res):
		self.res = res
		richstr = '<div align="center">'
		richstr += '<span style="font-size:10px;color:#990000;">Query of &lt; <strong><span style="color:#000099;">'
		richstr += str(sym)
		richstr += '</span></strong></span> <span style="font-size:10px;color:#990000;"> &gt; list</span>'
		richstr += '<div align="left" style="line-height:2px">'
		richstr += '<hr />'
		for itm in res:
			filename = itm[1]
			linenum = itm[2]
			context = itm[3]
			richstr += 'Line <span style="font-size:10px;color:#6600CC;">'
			richstr += str(linenum)
			richstr += '</span> of <em><a href="'
			richstr += str(filename)
			richstr += '#'
			richstr += str(linenum)
			richstr += '">'
			richstr += str(filename)
			richstr += '</a>'
			richstr += '</span></span></em>'
			richstr += '<pre style="line-height:0px"><span style="font-size:10px;color:#006600;line-height:2px;">'
			richstr += str(context)
			richstr += '</span></pre>'
		richstr += '</div></div>'
		#print richstr
		self.setHtml(richstr)
		self.setReadOnly(True)
		self.anchorClicked.connect(self.anchorClicked_ev)
		
	def anchorClicked_ev(self, qurl):
		urlstr = qurl.toString()
		urlinfo = urlstr.split('#')
		fname = str(urlinfo[0])
		linen = int(urlinfo[1])
		self.sig_result_view_openfile.emit(fname, int(linen))
	
	def setSource(self, qurl):
		self.setText('')
		self.anchorClicked.disconnect(self.anchorClicked_ev)

# contextView Page
class CodeContextViewPage(QFrame):
	def __init__(self, parent = None):
		QFrame.__init__(self)
		self.vlay = QVBoxLayout()

		self.cv_filetitle = QLabel()
		self.cv = ContextEditorView()
		self.resv = codeResultView()
		self.vlay.addWidget(self.cv_filetitle)
		self.vlay.addWidget(self.cv)
		self.vlay.addWidget(self.resv)
		#self.setSizes([1,1,1])
		self.resv.show()
		self.cv.hide()
		self.cv_filetitle.hide()

		self.setLayout(self.vlay)

	def showFileView(self, fname, line, sym, font):
		self.filename = fname
		self.linenum = line
		t = '%s:%s   %s' % (fname, line, sym)
		self.cv_filetitle.setText(t)
		self.cv_filetitle.show()

		cv = self.cv
		cv.set_lexer(fname)
		cv.set_font(font)
		cv.show_line_number_cb(True)
		cv.show_file_line(fname, line)
		cv.show()

# ContextView Manager
class CodeContextViewManager(QTabWidget):
	sig_codecontext_showfile = pyqtSignal(str, int)
	def __init__(self, *args):
		QTabWidget.__init__(*(self, ) + args)
		self.setMovable(True)
		self.setTabsClosable(True)
		self.newContextViewPage()
		self.cur_query = ''
		self.cvp_cv_openedFile = False
		
		self.cv_font = "Monospace,10,-1,5,50,0,0,0,0,0"

		self.cvp.cv.sig_dblclick.connect(self.contextViewPage_openfile)
		self.cvp.resv.sig_result_view_openfile.connect(self.resultViewPage_openfile)
	
	def change_ev_font(self, font):
		if font == self.cv_font:	
			return
		self.cv_font = font
		for inx in range(self.count()):
			cvp = self.widget(inx)
			cvp.cv.set_font(self.cv_font)

	def newContextViewPage(self):
		self.cvp = CodeContextViewPage()
		self.addTab(self.cvp, 'Context View')

	def set_cur_query(self, text):
		if self.cur_query == text:
			return False
		self.cur_query = text
		return True

	def showResult(self, sym, res):
		if sym != self.cur_query:
			return
		self.cvp.cv.clear()
			
		# if queried one, show the file, or, list results.
		if (len(res) == 1):
			self.cvp.resv.hide()
			itm = res[0]
			self.cvp.showFileView(itm[1], itm[2], sym, self.cv_font)
			self.cvp_cv_openedFile = True
		else:
			self.cvp.cv.hide()
			self.cvp.cv_filetitle.hide()
			self.cvp.resv.showResultList(sym, res)
			self.cvp.resv.show()
	
	def contextViewPage_openfile(self):
		self.cvp.cv.clear()
		self.sig_codecontext_showfile.emit(self.cvp.filename, int(self.cvp.linenum))
	
	def resultViewPage_openfile(self, filename, line):
		self.sig_codecontext_showfile.emit(filename, int(line))
			
	def clear(self):
		self.cvp.cv_filetitle.clear()
		self.cvp.resv.clear()
		self.cvp.cv.clear()
		self.cvp.cv.hide()



#!/usr/bin/env python

# Copyright (c) 2010 Anil Kumar
# All rights reserved.
#
# License: BSD 

import os
import re
import array

from PyQt6 import QtWidgets, QtGui
from PyQt6.QtWidgets import *
from PyQt6.QtGui import *
from PyQt6.QtCore import *

try:
	from PyQt6.Qsci import QsciScintilla, QsciScintillaBase
	from PyQt6.Qsci import QsciLexerCPP, QsciLexerJava
	from PyQt6.Qsci import QsciLexerPython, QsciLexerRuby
	from PyQt6.Qsci import QsciLexerBash, QsciLexerDiff, QsciLexerMakefile
	from PyQt6.Qsci import QsciLexerLua, QsciLexerSQL, QsciLexerTCL, QsciLexerTeX
	from PyQt6.Qsci import QsciLexerHTML, QsciLexerCSS
	from PyQt6.Qsci import QsciLexerPerl, QsciLexerVHDL

	suffix_to_lexer = [
		[['c', 'h', 'cpp', 'hpp', 'cc', 'hh', 'cxx', 'hxx', 'C', 'H', 'h++'], QsciLexerCPP],
		[['java'], QsciLexerJava],
		[['py', 'pyx', 'pxd', 'pxi', 'scons'], QsciLexerPython],
		[['rb', 'ruby'], QsciLexerRuby],
		[['sh', 'bash'], QsciLexerBash],
		[['diff', 'patch'], QsciLexerDiff],
		[['mak', 'mk'], QsciLexerMakefile],
		[['lua'], QsciLexerLua],
		[['sql'], QsciLexerSQL],
		[['tcl', 'tk', 'wish', 'itcl'], QsciLexerTCL],
		[['tex'], QsciLexerTeX],
		[['htm', 'html'], QsciLexerHTML],
		[['css'], QsciLexerCSS],
		[['pl', 'perl'], QsciLexerPerl],
		[['vhdl', 'vhd'], QsciLexerVHDL],
	]
	filename_to_lexer = [
		[['Makefile', 'makefile', 'Makefile.am', 'makefile.am', 'Makefile.in', 'makefile.in'], QsciLexerMakefile],
	]

	seascope_editor_tab_width = None
	try:
		seascope_editor_tab_width = int(os.getenv('SEASCOPE_EDITOR_TAB_WIDTH'))
	except:
		pass

except ImportError as e:
	print(e)
	print("Error: required python3-pyqt6.qsci package not found")
	raise ImportError

import DialogManager
from .FileContextView import *

class EditorViewBase(QsciScintilla):
	def __init__(self, parent=None):
		QsciScintilla.__init__(self, parent)
		self.font = None
		self.lexer = None

	def set_font(self, font):
		if not font:
			return
		if not self.font:
			self.font = QtGui.QFont()
		self.font.fromString(font)

		# the font metrics here will help
		# building the margin width later
		self.fm = QtGui.QFontMetrics(self.font)

		## set the default font of the editor
		## and take the same font for line numbers
		self.setFont(self.font)
		self.setMarginsFont(self.font)

		if self.lexer:
			# lexer can be null if no file is open at the moment
			self.lexer.setFont(self.font,-1)
			self.setLexer(self.lexer)

	def lpropChanged(self, prop, val):
		print('lpropChanged', prop, val)

	def setProperty(self, name, val):
		name_buff = array.array('b', (name + "\0").encode())
		val_buff = array.array("b", (str(val) + "\0").encode())
		address_name_buffer = name_buff.buffer_info()[0]
		address_val_buffer = val_buff.buffer_info()[0]
		self.SendScintilla(QsciScintillaBase.SCI_SETPROPERTY, address_name_buffer, address_val_buffer)

	def getProperty(self, name):
		name_buff = array.array('b', (name + "\0").encode())
		val_buff = array.array("b", (str(0) + "\0").encode())
		address_name_buffer = name_buff.buffer_info()[0]
		address_val_buffer = val_buff.buffer_info()[0]
		self.SendScintilla(QsciScintillaBase.SCI_GETPROPERTY, address_name_buffer, address_val_buffer)
		return ''.join(val_buff)

	def printPropertyAll(self):
		sz = self.SendScintilla(QsciScintillaBase.SCI_PROPERTYNAMES, 0, 0)
		if not sz:
			return
		val_buff = array.array("b", ((' ' * sz) + "\0").encode())
		address_val_buffer = val_buff.buffer_info()[0]
		self.SendScintilla(QsciScintillaBase.SCI_PROPERTYNAMES, 0, address_val_buffer)
		print('###>')
		for p in ''.join(val_buff).splitlines():
			v = self.getProperty(p)
			print('    %s = %s' % (p, v))

	def lexer_for_file(self, filename):
		(prefix, ext) = os.path.splitext(filename)
		for (el, lxr) in suffix_to_lexer:
			if ext in el:
				return lxr
		for (el, lxr) in filename_to_lexer:
			if filename in el:
				return lxr
		return QsciLexerCPP

	def set_lexer(self, filename):
		if not self.lexer:
			lexerClass = self.lexer_for_file(filename)
			self.lexer = lexerClass()
			#print('[%s: %s]' % (filename, self.lexer))
			self.setLexer(self.lexer)
			self.setProperty("lexer.cpp.track.preprocessor", "0")
			is_debug = os.getenv("SEASCOPE_QSCI_LEXER_DEBUG", 0)
			if is_debug:
				self.lexer.propertyChanged.connect(self.lpropChanged)
				self.printPropertyAll()


class EditorView(EditorViewBase):
	ev_popup = None
	sig_text_selected = pyqtSignal(str)
	def __init__(self, parent=None):
		EditorViewBase.__init__(self, parent)
		#self.setGeometry(300, 300, 400, 300)

		## Editing line color
		self.setCaretLineVisible(True)
		self.setCaretLineBackgroundColor(QtGui.QColor("#d4feff")) # orig: EEF6FF
		#self.setCaretWidth(2)

		self.setHorizontalScrollBarPolicy(Qt.ScrollBarPolicy.ScrollBarAsNeeded)

		self.codemark_marker = self.markerDefine(self.MarkerSymbol.Circle)

	def get_filename(self):
		return self.filename

	def ed_settings_1(self):
		## Margins colors
		# line numbers margin
		self.setMarginsBackgroundColor(QtGui.QColor("#333333"))
		self.setMarginsForegroundColor(QtGui.QColor("#CCCCCC"))

		# folding margin colors (foreground,background)
		self.setFoldMarginColors(QtGui.QColor("#888888"),QtGui.QColor("#eeeeee"))

		## Edge Mode shows a red vetical bar at 80 chars
		self.setEdgeMode(QsciScintilla.EdgeLine)
		self.setEdgeColumn(80)
		self.setEdgeColor(QtGui.QColor("#FF0000"))

		## Editing line color
		self.setCaretLineVisible(True)
		self.setCaretLineBackgroundColor(QtGui.QColor("#CDA869"))

		## set tab width
		if seascope_editor_tab_width:
			self.setTabWidth(seascope_editor_tab_width)

	def show_line_number_cb(self, val):
		if (val):
			width = self.fm.horizontalAdvance( "00000" ) + 5
		else:
			width = 0

		self.setMarginWidth(0, width)
		self.setMarginLineNumbers(0, val)

	def show_folds_cb(self, val):
		if val:
			#self.setMarginsForegroundColor( QtGui.QColor("#404040") )
			#self.setMarginsBackgroundColor( QtGui.QColor("#888888") )

			## Folding visual : we will use circled tree fold
			self.setFolding(QsciScintilla.FoldStyle.CircledTreeFoldStyle)
		else:
			self.setFolding(QsciScintilla.FoldStyle.NoFoldStyle)
			self.clearFolds()

	def toggle_folds_cb(self):
		self.foldAll()
		
	def codemark_add(self, line):
		self.markerAdd(line, self.codemark_marker)

	def codemark_del(self, line):
		self.markerDelete(line, self.codemark_marker)

	def goto_marker(self, is_next):
		(eline, inx) = self.getCursorPosition()
		u32max = 0xFFFFFFFF
		if is_next:
			val = self.markerFindNext(eline + 1, u32max)
		else:
			val = self.markerFindPrevious(eline - 1, u32max)
		if val >= 0:
			self.setCursorPosition(val, 0)

	def open_file_begin(self, filename):
		self.filename = filename

		## Choose a lexer
		self.set_lexer(filename)

		## Braces matching
		self.setBraceMatching(QsciScintilla.BraceMatch.SloppyBraceMatch)

		## Render on screen
		self.show()

	def open_file_end(self):
		self.show()
		self.setHorizontalScrollBarPolicy(Qt.ScrollBarPolicy.ScrollBarAsNeeded)
		#self.setHorizontalScrollBarPolicy(Qt.ScrollBarPolicy.ScrollBarAlwaysOff)
		self.setFocus()

	def open_file(self, filename):
		self.open_file_begin(filename)

		## Show this file in the editor
		self.setText(open(filename).read())
		
		## Mark read-only
		self.setReadOnly(True)

		self.open_file_end()

	def refresh_file(self, filename):
		assert filename == self.filename
		pos = self.getCursorPosition()
		self.open_file(filename)
		self.setCursorPosition(*pos)
		self.cursorPositionChanged.emit(*pos)

	def goto_line(self, line):
		line = line - 1
		self.setCursorPosition(line, 0)

		self.ensureLineVisible(line)
		self.setFocus()

	def contextMenuEvent(self, ev):
		if not EditorView.ev_popup:
			return
		f = EditorView.ev_popup.font()
		EditorView.ev_popup.setFont(QFont("San Serif", 8))
		EditorView.ev_popup.exec(QCursor.pos())
		EditorView.ev_popup.setFont(f)
		
	def mouseReleaseEvent(self, ev):
		super(EditorView, self).mouseReleaseEvent(ev)
		if(self.hasSelectedText()):
			self.query_text = self.selectedText()
			#print 'selectedText ----', self.query_text
			self.sig_text_selected.emit(self.query_text)

class EditorPage(QSplitter):
	def __init__(self, parent=None):
		QSplitter.__init__(self)
		self.fcv = FileContextView(self)
		self.ev = self.new_editor_view()
		self.addWidget(self.fcv)
		self.addWidget(self.ev)
		self.setSizes([1, 300])

		self.ev.cursorPositionChanged.connect(self.fcv.sig_ed_cursor_changed)
		self.fcv.sig_goto_line.connect(self.ev.goto_line)

	def new_editor_view(self):
		return EditorView(self)

	def open_file(self, filename):
		self.ev.open_file(filename)
		self.fcv.run(filename)

	def refresh_file(self):
		filename = self.get_filename()
		self.fcv.rerun(filename)
		self.ev.refresh_file(filename)

	def get_filename(self):
		return self.ev.get_filename()
		
class EditorBook(QTabWidget):
	sig_file_closed = pyqtSignal(str)
	sig_history_update = pyqtSignal(str, int)
	sig_tab_changed = pyqtSignal(str)
	sig_open_dir_view = pyqtSignal(str)
	sig_editor_text_selected = pyqtSignal(str)

	def __init__(self, *args):
		QTabWidget.__init__(*(self, ) + args)

		self.setMovable(True)
		self.setTabsClosable(True)
		self.tabCloseRequested.connect(self.removeTab)
		self.currentChanged.connect(self.tab_change_cb)

		self.is_show_line = False
		self.is_show_folds = False
		self.f_text = None
		self.ev_font = "Monospace,10,-1,5,50,0,0,0,0,0"

	def new_editor_page(self):
		return EditorPage()

	def addFile(self, fileName):
		ed = self.new_editor_page()
		ed.open_file(fileName)
		ed.ev.set_font(self.ev_font)
		ed.ev.show_line_number_cb(self.is_show_line)
		ed.ev.show_folds_cb(self.is_show_folds)
		self.addTab(ed, os.path.basename(fileName))
		return ed

	def search_already_opened_files(self, filename):
		for i in range(self.count()):
			page = self.widget(i)
			if (page.get_filename() == filename):
				return page
		return None

	def removeTab(self, inx):
		ed = self.widget(inx)
		if not ed:
		    return
		f = ed.ev.get_filename()
		QTabWidget.removeTab(self, inx)
		self.sig_file_closed.emit(f)

	def clear(self):
		while self.count() != 0:
			self.removeTab(0)

	def remove_tab_list(self, inx_list):
		for inx in sorted(inx_list, reverse=True):
			self.removeTab(inx)

	def tab_list(self, inx, type):
		inx_list = []
		if type == 'all' or type == 'files':
			return list(range(self.count()))
		if type == 'left':
			return list(range(inx))
		if type == 'right':
			return list(range(inx + 1, self.count()))
		if type == 'other':
			return self.tab_list(inx, 'left') + self.tab_list(inx, 'right')
		assert 0

	def close_list_common(self, type):
		inx_list = self.tab_list(self.currentIndex(), type)
		if len(inx_list) == 0:
			return
		if not DialogManager.show_yes_no('Close all %s ?' % type):
			return
		self.remove_tab_list(inx_list)

	def close_all_left_cb(self):
		self.close_list_common('left')
	def close_all_right_cb(self):
		self.close_list_common('right')
	def close_all_other_cb(self):
		self.close_list_common('other')
	def close_all_cb(self):
		self.close_list_common('files')

	def get_current_word(self):
		ed = self.currentWidget()
		if not ed:
			return
		if ed.ev.hasSelectedText():
			return ed.ev.selectedText()
		(line, index) = ed.ev.getCursorPosition()
		text = ed.ev.text(line)
		# Go left
		linx = index
		while linx > 0 and ed.ev.isWordCharacter(text[linx - 1]):
			linx = linx - 1
		# Go right
		rinx = index
		while rinx < len(text) and ed.ev.isWordCharacter(text[rinx]):
			rinx = rinx + 1
		
		text = text[linx:rinx]
		if text == '':
			return None
		return text
	def get_current_file_line(self):
		ed = self.currentWidget()
		if not ed:
			return (None, None)
		(line, inx) = ed.ev.getCursorPosition()
		return (ed.ev.filename, line + 1)
	def get_file_line_list(self):
		fl_list = []
		tlist = list(range(self.count()))
		inx = self.currentIndex()
		if inx >= 0:
			tlist.append(inx)
		for inx in tlist:
			ed = self.widget(inx)
			(line, inx) = ed.ev.getCursorPosition()
			fl_list.append('%s:%d' % (ed.ev.filename, line + 1))
		return fl_list

	def matching_brace_cb(self):
		ed = self.currentWidget()
		if ed:
			ed.ev.moveToMatchingBrace()
			#ed.ev.setFocus()
	def goto_line_cb(self):
		ed = self.currentWidget()
		if not ed:
			return (None, None)
		(line, inx) = ed.ev.getCursorPosition()
		#return (line + 1, ed.ev.lines())
		line = DialogManager.show_goto_line_dialog(line + 1, ed.ev.lines())
		if (line == None):
			return
		ed.ev.goto_line(line)	

	def focus_editor(self):
		page = self.currentWidget()
		if page:
			page.ev.setFocus()

	def close_current_page(self):
		self.removeTab(self.currentIndex())
		self.focus_editor()

	def focus_search_ctags(self):
		ed = self.currentWidget()
		if ed:
			ed.fcv.focus_search_ctags()
		
	def copy_edit_cb(self):
		ed = self.currentWidget()
		if ed:
			ed.ev.copy()

	def tab_change_cb(self, inx):
		if (inx == -1):
			fname = ''
		else:
			page = self.currentWidget()
			page.ev.setFocus()
			fname = page.get_filename()
		self.sig_tab_changed.emit(fname)
			
	def open_dir_cb(self):
		page = self.currentWidget()
		if page:
			fname = page.get_filename()
			self.sig_open_dir_view.emit(fname)
			

	def mousePressEvent(self, m_ev):
		QTabWidget.mousePressEvent(self, m_ev)
		if (m_ev.button() == Qt.MouseButton.RightButton):
			# setup popup menu
			self.pmenu = QMenu()
			self.pmenu.addAction("Open dir", self.open_dir_cb)
			self.pmenu.addSeparator()
			self.pmenu.addAction("Close All &Left", self.close_all_left_cb)
			self.pmenu.addAction("Close All &Right", self.close_all_right_cb)
			self.pmenu.addAction("Close &Others", self.close_all_other_cb)
			self.pmenu.addSeparator()
			self.pmenu.addAction("Close &All", self.close_all_cb)
			self.pmenu.exec(QCursor.pos())

	def show_file_line(self, filename, line, hist=True):
		if line:
			(f, l) = self.get_current_file_line()
			if (f):
				if hist:
					self.sig_history_update.emit(f, l)
		filename = str(filename)
		if (not os.path.exists(filename)):
			return
		page = self.search_already_opened_files(filename)
		if page == None:
			page = self.addFile(filename)
		self.setCurrentWidget(page)
		if line:
			page.ev.goto_line(line)
			if hist:
				self.sig_history_update.emit(filename, line)
		page.ev.setFocus()
		# text selected callback: need to send out again.
		page.ev.sig_text_selected.connect(self.editor_text_selected)

	def editor_text_selected(self, text):
		self.sig_editor_text_selected.emit(text)

	def show_file(self, filename):
		self.show_file_line(filename, 0)

	def show_line(self, line):
		ed = self.currentWidget()
		if not ed:
			return
		ed.ev.goto_line(line)

	def refresh_file_cb(self):
		ed = self.currentWidget()
		if not ed:
			return
		ed.refresh_file()

	def find_cb(self):
		ed = self.currentWidget()
		if not ed:
			return
		res = DialogManager.show_find_dialog(self.get_current_word())
		if (res == None):
			return
		(text, opt) = res
		if (text == None):
			return
		self.f_text = text
		self.f_opt = opt
		self.find_text(opt['cursor'], opt['fw'])

	def find_text(self, from_cursor, is_fw):
		if (self.f_text == None):
			return
		ed = self.currentWidget()
		if not ed:
			return
		text = self.f_text
		opt = self.f_opt
		if (from_cursor):
			if (is_fw):
				(line, inx) = (-1, -1)
			else:
				(line, inx) = ed.ev.getCursorPosition()
				if (ed.ev.hasSelectedText()):
					inx = inx - 1
			if ed.ev.findFirst(text, opt['re'], opt['cs'], opt['wo'], False, is_fw, line, inx):
				return True
			if not DialogManager.show_yes_no('End of document reached. Continue from beginning?'):
				return False
		if (is_fw):
			(line, inx) = (0, 0)
		else:
			(line, inx) = (ed.ev.lines(), 0)
		if ed.ev.findFirst(text, opt['re'], opt['cs'], opt['wo'], False, is_fw, line, inx):
			return
		DialogManager.show_msg_dialog("Could not find " + "'" + text + "'")

	def find_next_cb(self):
		self.find_text(True, True)

	def find_prev_cb(self):
		self.find_text(True, False)

	def change_ev_font(self, font):
		if font == self.ev_font:	
			return
		self.ev_font = font
		for inx in range(self.count()):
			ed = self.widget(inx)
			ed.ev.set_font(self.ev_font)

	def show_line_number_cb(self):
		val = self.m_show_line_num.isChecked()
		self.is_show_line = val
		for inx in range(self.count()):
			ed = self.widget(inx)
			ed.ev.show_line_number_cb(val)

	def show_line_number_pref(self, val):
		if val == self.m_show_line_num.isChecked():
			return
		self.m_show_line_num.setChecked(val)
		self.show_line_number_cb()

	def show_folds_cb(self):
		val = self.m_show_folds.isChecked()
		self.is_show_folds = val
		for inx in range(self.count()):
			ed = self.widget(inx)
			ed.ev.show_folds_cb(val)

	def toggle_folds_cb(self):
		ed = self.currentWidget()
		if not ed:
			return
		if self.is_show_folds:
			ed.ev.toggle_folds_cb()

	def open_in_external_editor(self, cmd):
		if not cmd:
			DialogManager.show_msg_dialog('Please configure external editor')
			return
		(f, l) = self.get_current_file_line()
		if not f:
			return
		cmd = cmd.replace('%F', f).replace('%L', str(l))
		if not QProcess.startDetached(cmd):
			DialogManager.show_msg_dialog('Failed to start: ' + cmd)

	def codemark_add(self, filename, line):
		ed = self.search_already_opened_files(filename)
		if ed:
			ed.ev.codemark_add(line)

	def codemark_del(self, filename, line):
		ed = self.search_already_opened_files(filename)
		if ed:
			ed.ev.codemark_del(line)

	def bookmark_prev_cb(self):
		ed = self.currentWidget()
		if ed:
			ed.ev.goto_marker(False)

	def bookmark_next_cb(self):
		ed = self.currentWidget()
		if ed:
			ed.ev.goto_marker(True)



# Copyright (c) 2010 Anil Kumar
# All rights reserved.
#
# License: BSD 

import os

def name():
	return 'cscope'

def is_your_prj(path):
	f = os.path.join(path, 'cscope.files')
	return os.path.exists(f)

def project_class():
	from .CscopeProject import ProjectCscope
	return ProjectCscope

def description():
	d = 'Cscope supports C, lex, yacc files.\n'
	d += 'Support for C++ and Java is limited.'
	return d

priority = 500





#!/usr/bin/env python3

# Copyright (c) 2010 Anil Kumar
# All rights reserved.
#
# License: BSD 

import os, string, re

from ..PluginBase import PluginFeatureBase, ProjectBase, ConfigBase, QueryBase
from ..PluginBase import PluginProcess


class CscopeFeature(PluginFeatureBase):
	def __init__(self):
		PluginFeatureBase.__init__(self)

		self.feat_desc = [
			['REF',	'0'],
			['DEF',	'1'],
			['<--',	'2'],
			['-->',	'3'],
			['TXT',	'4'],
			['GREP','5'],
			['FIL',	'7'],
			['INC',	'8'],

			['QDEF', '11'],
			['CTREE','12'],

			['CLGRAPH', '13'],
			['CLGRAPHD', '14'],
			['FFGRAPH', '14'],

			['UPD', '25'],
		]

		self.ctree_query_args = [
			['-->',	'--> F', 'Calling tree'			],
			['<--',	'F -->', 'Called tree'			],
			['REF',	'==> F', 'Advanced calling tree'	],
		]
				
	def query_dlg_cb(self, req, cmd_str, in_opt):
		if req != '' and in_opt['substring']:
			req = '.*' + req + '.*'
		opt = None
		if in_opt['ignorecase']:
			opt = '-C'
		res = (cmd_str, req, opt)
		return res

class ConfigCscope(ConfigBase):
	def __init__(self):
		ConfigBase.__init__(self, 'cscope')
		self.c_opt = []

	def read_cs_files_common(self, filename):
		fl = []
		config_file = os.path.join(self.c_dir, filename)
		if (os.path.exists(config_file)):
			cf = open(config_file, 'r')
			for f in cf:
				f = f[:-1]
				fl.append(f)
			cf.close()
		return fl

	def read_cs_files(self):
		self.c_flist = self.read_cs_files_common('cscope.files')

	def write_cs_files_common(self, filename, fl):
		if (len(fl) <= 0):
			return
		config_file = os.path.join(self.c_dir, filename)
		cf = open(config_file, 'w')
		for f in fl:
			cf.write(f + '\n')
		cf.close()

	def write_cs_files(self):
		self.write_cs_files_common("cscope.files", self.c_flist)

	def get_config_file(self):
		config_file = 'seascope.opt'
		return os.path.join(self.c_dir, config_file)

	def read_seascope_opt(self):
		config_file = self.get_config_file()
		if (not os.path.exists(config_file)):
			return
		cf = open(config_file, 'r')
		for line in cf:
			line = line.split('=', 1)
			key = line[0].strip()
			if (key == 'c_opt'):
				self.c_opt = line[1].split()
		cf.close()
		
	def write_seascope_opt(self):
		config_file = self.get_config_file()
		cf = open(config_file, 'w')
		cf.write('c_opt' + '=' + ' '.join(self.c_opt) + '\n')
		cf.close()
		
	def read_config(self):
		self.read_seascope_opt()
		self.read_cs_files()

	def write_config(self):
		self.write_seascope_opt()
		self.write_cs_files()

	def is_ready(self):
		return len(self.c_flist) > 0

class ProjectCscope(ProjectBase):
	def __init__(self):
		ProjectBase.__init__(self)

	@staticmethod
	def _prj_new_or_open(conf):
		prj = ProjectCscope()
		prj.feat = CscopeFeature()
		prj.conf = conf
		prj.qry = QueryCscope(prj.conf, prj.feat)
		
		return (prj)

	@staticmethod
	def prj_new(proj_args):
		conf = ConfigCscope()
		conf.proj_new(proj_args)
		prj = ProjectCscope._prj_new_or_open(conf)
		return (prj)

	@staticmethod
	def prj_open(proj_path):
		conf = ConfigCscope()
		conf.proj_open(proj_path)
		prj = ProjectCscope._prj_new_or_open(conf)
		return (prj)

	def prj_settings_update(self, proj_args):
		assert proj_args
		self.prj_update_conf(proj_args)
		return True


class CsProcess(PluginProcess):
	def __init__(self, wdir, rq):
		PluginProcess.__init__(self, wdir, rq)
		self.name = 'cscope process'

	def parse_result(self, text, sig):
		text = re.split(r'\r?\n', text)
		if self.cmd_str == 'FIL':
			res = [ ['', line[0], '', ''] for line in text if line != '' ]
			return res
		res = []
		for line in text:
			if line == '':
				continue
			line = line.split(' ', 3)
			line = [line[1], line[0], line[2], line[3]]
			res.append(line)
		return res

class QueryCscope(QueryBase):
	def __init__(self, conf, feat):
		QueryBase.__init__(self)
		self.conf = conf
		self.feat = feat

	def query(self, rquery):
		if (not self.conf or not self.conf.is_ready()):
			print("pm_query not is_ready")
			return None
		cmd_str = rquery['cmd']
		req     = rquery['req']
		opt     = rquery['opt']
		cmd_id = self.feat.cmd_str2id[cmd_str]
		if opt == None or opt == '':
			opt = []
		else:
			opt = opt.split()
		pargs  = [ 'cscope' ] + self.conf.c_opt + opt + [ '-L', '-d',  '-' + str(cmd_id), req ]
		qsig = CsProcess(self.conf.c_dir, [cmd_str, req]).run_query_process(pargs, req, rquery)
		return qsig

	def rebuild(self):
		if (not self.conf.is_ready()):
			print("pm_query not is_ready")
			return None
		pargs = [ 'cscope' ] + self.conf.c_opt + [ '-L' ]
		qsig = CsProcess(self.conf.c_dir, None).run_rebuild_process(pargs)
		return qsig

	def query_fl(self):
		fl = self.conf.get_proj_src_files()
		return fl

	def cs_is_open(self):
		return self.conf != None
	def cs_is_ready(self):
		return self.conf.is_ready()



# Copyright (c) 2010 Anil Kumar
# All rights reserved.
#
# License: BSD 

import os
import sys
import re

from PyQt6 import QtWidgets, QtGui, QtCore, uic

from PyQt6.QtWidgets import *
from PyQt6.QtGui import *
from PyQt6.QtCore import *

def show_msg_dialog(msg):
	QMessageBox.warning(None, "Seascope", msg, QMessageBox.StandardButton.Ok)

def dir_scan_csope_files(rootdir):
	file_list = []
	if (not os.path.isdir(rootdir)):
		print("Not a directory:", rootdir)
		return file_list
	for root, subFolders, files in os.walk(rootdir):
		for f in files:
			f = os.path.join(root, f)
			if (re.search(r'\.(h|c|H|C|hh|cc|hpp|cpp|hxx|cxx||l|y|s|S|pl|pm|java)$', f) != None):
				file_list.append(f)
	return file_list

class CscopeProjectSettingsDialog(QDialog):
	def __init__(self):
		QDialog.__init__(self)

		self.ui = uic.loadUi('ui/cscope_prj_settings.ui', self)
		self.pd_path_tbtn.setIcon(QFileIconProvider().icon(QFileIconProvider.IconType.Folder))

		self.pd_src_list.setSelectionMode(QAbstractItemView.SelectionMode.ExtendedSelection)

		self.pd_path_tbtn.clicked.connect(self.path_open_cb)
		self.pd_add_btn.clicked.connect(self.src_add_cb)
		self.pd_rem_btn.clicked.connect(self.src_rem_cb)
		self.accepted.connect(self.ok_btn_cb)

	def path_open_cb(self):
		fdlg = QFileDialog(None, "Choose directory")
		fdlg.setFileMode(QFileDialog.FileMode.Directory);
		fdlg.setDirectory(self.pd_path_inp.text())
		if (fdlg.exec()):
			path_dir = fdlg.selectedFiles()[0];
			self.pd_path_inp.setText(str(path_dir))

	def src_add_cb(self):
		fdlg = QFileDialog(None, "Choose directory")
		fdlg.setFileMode(QFileDialog.FileMode.Directory);
		fdlg.setDirectory(self.pd_path_inp.text())
		if (fdlg.exec()):
			d = fdlg.selectedFiles()[0];
			d = str(d)
			d = d.strip()
			if (d == None):
				return
			d = os.path.normpath(d)
			if (d == '' or not os.path.isabs(d)):
				return
			self.src_add_files(d)

	def src_rem_cb(self):
		li = self.pd_src_list
		for item in li.selectedItems():
			row = li.row(item)
			li.takeItem(row)

	def src_add_files(self, src_dir):
		file_list = dir_scan_csope_files(src_dir)
		self.pd_src_list.addItems(file_list)

	def ok_btn_cb(self):
		proj_dir = os.path.join(str(self.pd_path_inp.text()), str(self.pd_name_inp.text()))
		proj_dir = os.path.normpath(proj_dir)
		if (self.is_new_proj):
			if (proj_dir == '' or not os.path.isabs(proj_dir)):
				return
			if (os.path.exists(proj_dir)):
				show_msg_dialog("\nProject already exists")
				return
			os.mkdir(proj_dir)
		# File list
		cs_list = []
		for inx in range(self.pd_src_list.count()):
			val = str(self.pd_src_list.item(inx).text())
			cs_list.append(val)
		cs_list = list(set(cs_list))
		# Cscope opt
		cs_opt = []
		if self.pd_invert_chkbox.isChecked():
			cs_opt.append('-q')
		if self.pd_kernel_chkbox.isChecked():
			cs_opt.append('-k')

		self.res = [proj_dir, cs_opt, cs_list]

	def set_proj_args(self, proj_args):
		(proj_dir, cs_opt, cs_list) = proj_args
		(proj_base, proj_name) = os.path.split(proj_dir)
		self.pd_path_inp.setText(proj_base)
		self.pd_name_inp.setText(proj_name)
		# File list
		fl = cs_list
		self.pd_src_list.addItems(fl)
		# Cscope opt
		for opt in cs_opt:
			if (opt == '-q'):
				self.pd_invert_chkbox.setChecked(True)
			if (opt == '-k'):
				self.pd_kernel_chkbox.setChecked(True)

	def run_dialog(self, prj_type, proj_args):
		assert prj_type == 'cscope'
		self.pd_src_list.clear()
		if (proj_args == None):
			self.is_new_proj = True
			self.pd_invert_chkbox.setChecked(True)
		else:
			self.is_new_proj = False
			self.set_proj_args(proj_args)
		self.res = None
		
		self.pd_path_frame.setEnabled(self.is_new_proj)

		while True:
			ret = self.exec()
			if (ret == QDialog.DialogCode.Accepted or ret == QDialog.DialogCode.Rejected):
				break
		return self.res



#!/usr/bin/env python

# Copyright (c) 2010 Anil Kumar
# All rights reserved.
#
# License: BSD 

import sys
import os
import string

from PyQt6 import QtWidgets, QtGui, QtCore, uic

from PyQt6.QtWidgets import *
from PyQt6.QtGui import *
from PyQt6.QtCore import *


def show_msg_dialog(msg):
	QMessageBox.information(None, "Seascope", msg, QMessageBox.StandardButton.Ok)

def show_yes_no(msg):
	ret = QMessageBox.question(None, "Seascope", msg, QMessageBox.StandardButton.Yes | QMessageBox.StandardButton.No, QMessageBox.StandardButton.Yes)
	return ret == QMessageBox.StandardButton.Yes

def show_yes_no_dontask(msg):
	mb = QMessageBox(None)
	mb.setIcon(QMessageBox.Icon.Question)
	mb.setWindowTitle("Seascope")
	mb.setText(msg)
	mb.addButton('Yes', QMessageBox.ButtonRole.YesRole)
	mb.addButton('No', QMessageBox.ButtonRole.NoRole)
	mb.addButton("Yes, Don't ask again", QMessageBox.ButtonRole.YesRole)
	return mb.exec()

def show_proj_close():
	return show_yes_no("\nClose current project?")

class ProjectOpenDialog(QObject):
	def __init__(self):
		QObject.__init__(self)

		self.dlg = uic.loadUi('ui/proj_open.ui')
		self.dlg.pod_open_btn.setIcon(QFileIconProvider().icon(QFileIconProvider.IconType.Folder))

		self.dlg.pod_open_btn.clicked.connect(self.open_btn_cb)
		self.dlg.accepted.connect(self.ok_btn_cb)
		self.dlg.pod_proj_list.itemSelectionChanged.connect(self.proj_list_change_cb)

	def proj_list_change_cb(self):
		item = self.dlg.pod_proj_list.selectedItems()[0]
		proj_dir = str(item.text())
		self.dlg.pod_proj_name.setText(str(proj_dir))
		self.path = proj_dir

	def open_btn_cb(self):
		fdlg = QFileDialog(None, "Choose project directory")
		fdlg.setFileMode(QFileDialog.FileMode.Directory);
		fdlg.setDirectory(self.dlg.pod_proj_name.text())
		if (fdlg.exec()):
			proj_dir = fdlg.selectedFiles()[0];
			self.dlg.pod_proj_name.setText(str(proj_dir))
		
	def ok_btn_cb(self):
		p = str(self.dlg.pod_proj_name.text()).strip()
		if (p == ''):
			self.dlg.setResult(2)
			return
		p = os.path.normpath(p)
		if (not os.path.isabs(p)):
			show_msg_dialog("\nProject path is not absolute")
			self.dlg.setResult(2)
			return
		if (not os.path.isdir(p)):
			show_msg_dialog("\nProject path is not a directory")
			self.dlg.setResult(2)
			return
		self.path = p

	def run_dialog(self, path_list):
		self.path = None
		self.dlg.pod_proj_name.setText('')
		self.dlg.pod_proj_list.addItems(path_list)
		while True:
			ret = self.dlg.exec()
			if (ret == QDialog.DialogCode.Accepted or ret == QDialog.DialogCode.Rejected):
				break
		return self.path


def show_project_open_dialog(path_list):
	d = ProjectOpenDialog()
	return d.run_dialog(path_list)

class FilePreferencesDialog(QObject):
	def __init__(self, app_style, edit_ext_cmd, ev_font, dontask, innered, ln_nr):
		QObject.__init__(self)

		self.dlg = uic.loadUi('ui/preferences.ui')
		self.dlg.prd_style_lw.addItems(list(QStyleFactory.keys()))
		self.dlg.prd_style_lw.itemSelectionChanged.connect(self.style_changed_cb)
		self.dlg.prd_font_app_btn.clicked.connect(self.font_app_btn_cb)
		self.dlg.prd_font_ev_btn.clicked.connect(self.font_ev_btn_cb)
		self.set_btn_text_and_font(self.dlg.prd_font_app_btn, QApplication.font())
		self.set_btn_text_and_font(self.dlg.prd_font_ev_btn, ev_font)
		self.app_style = app_style
		self.ev_font = ev_font
		self.edit_ext_cmd = edit_ext_cmd
		self.exit_dontask = dontask
		self.inner_editing = innered
		self.show_ln_nr = ln_nr
		if self.exit_dontask:
			self.dlg.prd_opt_ask_chkb.setCheckState(Qt.CheckState.Unchecked)
		else:
			self.dlg.prd_opt_ask_chkb.setCheckState(Qt.CheckState.Checked)
		if self.inner_editing:
			self.dlg.prd_opt_inner_ed.setCheckState(Qt.CheckState.Checked)
		else:
			self.dlg.prd_opt_inner_ed.setCheckState(Qt.CheckState.Unchecked)
		if self.show_ln_nr:
			self.dlg.prd_opt_show_ln_nr.setCheckState(Qt.CheckState.Checked)
		else:
			self.dlg.prd_opt_show_ln_nr.setCheckState(Qt.CheckState.Unchecked)
		if (self.edit_ext_cmd):
			self.dlg.prd_edit_ext_inp.setText(self.edit_ext_cmd)

	def set_btn_text_and_font(self, btn, font):
		ftext = font.family() + ' ' + str(font.pointSize())
		btn.setFont(font)
		btn.setText(ftext)

	def style_changed_cb(self):
		item = self.dlg.prd_style_lw.currentItem()
		self.app_style = item.text()
		QApplication.setStyle(self.app_style)

	def font_app_btn_cb(self):
		(font, ok) = QFontDialog().getFont(QApplication.font())
		if (ok):
			self.set_btn_text_and_font(self.dlg.prd_font_app_btn, font)

	def font_ev_btn_cb(self):
		(font, ok) = QFontDialog().getFont(self.ev_font)
		if (ok):
			self.set_btn_text_and_font(self.dlg.prd_font_ev_btn, font)
			
	def run_dialog(self):
		ret = self.dlg.exec()
		if (ret == QDialog.DialogCode.Accepted):
			QApplication.setFont(self.dlg.prd_font_app_btn.font())
			self.ev_font = self.dlg.prd_font_ev_btn.font()
			self.edit_ext_cmd = self.dlg.prd_edit_ext_inp.text()
			self.exit_dontask = self.dlg.prd_opt_ask_chkb.checkState() == Qt.CheckState.Unchecked
			self.inner_editing = self.dlg.prd_opt_inner_ed.checkState() == Qt.CheckState.Checked
			self.show_ln_nr = self.dlg.prd_opt_show_ln_nr.checkState() == Qt.CheckState.Checked
		return (self.app_style, self.dlg.prd_font_app_btn.font().toString(), self.edit_ext_cmd, 
			self.ev_font.toString(), self.exit_dontask, self.inner_editing, self.show_ln_nr)

def show_preferences_dialog(app_style, edit_ext_cmd, ev_font, dontask, innered, ln_nr):
	d = FilePreferencesDialog(app_style, edit_ext_cmd, ev_font, dontask, innered, ln_nr)
	return d.run_dialog()

def show_about_dialog():
	d = uic.loadUi('ui/about.ui')
	p = QApplication.windowIcon().pixmap(128)
	d.ad_logo_lbl.setPixmap(p)
	d.exec()

def show_goto_line_dialog(line, max_line):
	d = uic.loadUi('ui/goto_line.ui')
	d.gl_spinbox.setMaximum(max_line)
	d.gl_hslider.setMaximum(max_line)
	d.gl_spinbox.setValue(line)
	d.gl_spinbox.lineEdit().selectAll()
	d.gl_spinbox.lineEdit().setFocus()
	if (d.exec() == QDialog.DialogCode.Accepted):
		return d.gl_spinbox.value()
	return None

class FindDialog:
	dlg = None
	def __init__(self):
		self.dlg = uic.loadUi('ui/find.ui')
		self.dlg.ft_text_inp.setCompleter(None)
		self.dlg.ft_text_inp.setInsertPolicy(QComboBox.InsertPolicy.InsertAtTop)
		self.dlg.ft_text_inp.setFocus()

	def run_dlg(self, text):
		if (text == None):
			text = ''
		d = self.dlg
		d.ft_text_inp.lineEdit().setText(text)
		d.ft_text_inp.lineEdit().selectAll()
		d.ft_text_inp.lineEdit().setFocus()
		d.ft_from_cursor.setChecked(True)
		while True:
			ret = d.exec()
			if (ret != QDialog.DialogCode.Accepted):
				return None
			text = d.ft_text_inp.currentText()
			text = str(text).strip()
			if (text != ''):
				break
		opt = {
			'cs' : d.ft_cs.isChecked(),
			're' : d.ft_re.isChecked(),
			'wo' : d.ft_wo.isChecked(),
			'fw' : not d.ft_bw.isChecked(),
			'cursor' : d.ft_from_cursor.isChecked(),
		}
		return (text, opt)

def show_find_dialog(text):
	if (FindDialog.dlg == None):
		FindDialog.dlg = FindDialog()
	return FindDialog.dlg.run_dlg(text)

class FilterDialog:
	def __init__(self, cmd_items, cmd_sel):
		self.dlg = uic.loadUi('ui/filter.ui')
		self.dlg.fd_cmd_inp.addItems(cmd_items)
		self.dlg.fd_cmd_inp.setCurrentIndex(cmd_items.index(cmd_sel))
		self.dlg.fd_filter_inp.setFocus()
		self.dlg.fd_regex_err_lbl.setVisible(False)

	def run_dialog(self):
		while True:
			ret = self.dlg.exec()
			if ret != QDialog.DialogCode.Accepted:
				return None
			is_regex = self.dlg.fd_regex_chkbox.isChecked()
			text = str(self.dlg.fd_filter_inp.text())
			if is_regex == True:
				try:
					import re
					re.compile(text)
				except:
					e = str(sys.exc_info()[1])
					lbl = self.dlg.fd_regex_err_lbl
					lbl.setVisible(True)
					lbl.setText("<font color='red'>regex: " + e + '</font>')
					continue
					
			res = [
				text,
				is_regex,
				self.dlg.fd_negate_chkbox.isChecked(),
				self.dlg.fd_icase_chkbox.isChecked(),
				str(self.dlg.fd_cmd_inp.currentText())
			]
			return res

def show_filter_dialog(cmd_items, cmd_sel):
	dlg = FilterDialog(cmd_items,cmd_sel)
	return dlg.run_dialog()



# Copyright (c) 2010 Anil Kumar
# All rights reserved.
#
# License: BSD 

import os

def name():
	return 'gtags'

def is_your_prj(path):
	f = os.path.join(path, 'GTAGS')
	return os.path.exists(f)

def project_class():
	from .GtagsProject import ProjectGtags
	return ProjectGtags

def description():
	d = 'GNU global/gtags supports C, C++, Yacc, Java, PHP and Assembly source files.\n'
	d += '\nSeascope uses ctags to enhance the results.'
	return d

#priority = 200



# Copyright (c) 2010 Anil Kumar
# All rights reserved.
#
# License: BSD 

import os

def name():
	return 'ctags'

def run_plugin(filename, parent):
	from . import CtagsView
	CtagsView.run(filename, parent)

def description():
	d = 'ctags description'
	return d

priority = 500



#!/usr/bin/env python

# Copyright (c) 2014 Anil Kumar
# All rights reserved.
#
# License: BSD

import sys
import os
import subprocess
import threading

class PluginProcess:
	def __init__(self, pargs):
		self.pargs = pargs
		self.proc = None
		self.res = {
				'out_data' : None,
				'err_data' : None,
			}

	def run(self):
		try:
			self.proc = subprocess.Popen(self.pargs, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
			(out_data, err_data) = self.proc.communicate()
			self.res['out_data'] = out_data.decode()
			self.res['err_data'] = err_data.decode()
		except Exception as e:
			print('PluginProcess.run():', e)
			self.res['err_data'] = str(e)
		return self.res



#!/usr/bin/env python3

# Copyright (c) 2010 Anil Kumar
# All rights reserved.
#
# License: BSD 

import sys, os, string, re

from ..PluginBase import PluginFeatureBase, ProjectBase, ConfigBase, QueryBase
from ..PluginBase import PluginProcess
from ..CtagsCache import CtagsThread


class IdutilsFeature(PluginFeatureBase):
	def __init__(self):
		PluginFeatureBase.__init__(self)

		self.feat_desc = [
			['REF',     ''],
			['DEF',     ''],
			#['<--',    '2'],
			['-->',     '3'],
			#['TXT',    '4'],
			['GREP',    ''],
			['FIL',     ''],
			['INC',     '8'],

			['QDEF',    ''],
			['CTREE',   '12'],

			['CLGRAPH', '13'],
			['CLGRAPHD','14'],
			['FFGRAPH', '14'],

			['UPD',    '25'],
		]

		self.ctree_query_args = [
			['-->',	'--> F', 'Calling tree'			],
			#['<--',	'F -->', 'Called tree'			],
			['REF',	'==> F', 'Advanced calling tree'	],
		]

	def query_dlg_cb(self, req, cmd_str, in_opt):
		opt = []
		if cmd_str != 'TXT' and req != '' and in_opt['substring']:
			opt.append('substring')
			if cmd_str == 'FIL':
				req = '*' + req + '*'
			else:
				req = '.*' + req + '.*'
		if in_opt['ignorecase']:
			opt.append('ignorecase')

		res = (cmd_str, req, opt)
		return res



class ConfigIdutils(ConfigBase):
	def __init__(self):
		ConfigBase.__init__(self, 'idutils')

class ProjectIdutils(ProjectBase):
	def __init__(self):
		ProjectBase.__init__(self)

	@staticmethod
	def _prj_new_or_open(conf):
		prj = ProjectIdutils()
		prj.feat = IdutilsFeature()
		prj.conf = conf
		prj.qry = QueryIdutils(prj.conf, prj.feat)
		return (prj)

	@staticmethod
	def prj_new(proj_args):
		d = proj_args[0]
		prj = ProjectIdutils.prj_open(d)
		return prj

	@staticmethod
	def prj_open(proj_path):
		conf = ConfigIdutils()
		conf.proj_open(proj_path)
		prj = ProjectIdutils._prj_new_or_open(conf)
		return (prj)


class IdProcess(PluginProcess):
	def __init__(self, wdir, rq):
		PluginProcess.__init__(self, wdir, rq)
		self.name = 'idutils process'

	def parse_result(self, text, sig):
		if self.cmd_str in ['GREP']:
			return PluginProcess.parse_result(self, text, sig)

		#from datetime import datetime
		#t1 = datetime.now()

		text = re.split(r'\r?\n', text)

		#t2 = datetime.now()
		#print 'parse-split', t2 - t1

		if self.cmd_str == 'FIL':
			res = [ ['',  os.path.join(self.wdir, line), '', '' ] for line in text if line != '' ]
			return res

		res = []
		for line in text:
			if line == '':
				continue
			line = line.split(':', 2)
			line = ['<unknown>', os.path.join(self.wdir, line[0]), line[1], line[2]]
			res.append(line)

		#t3 = datetime.now()
		#print 'parse-loop', t3 - t2

		res = CtagsThread(sig).apply_fix(self.cmd_str, res, ['<unknown>'])

		return res

idutils_lid_cmd = os.getenv('SEASCOPE_IDUTILS_LID_CMD', 'lid')

class QueryIdutils(QueryBase):
	def __init__(self, conf, feat):
		QueryBase.__init__(self)
		self.conf = conf
		self.feat = feat

	def query(self, rquery):
		if (not self.conf):
		#or not self.conf.is_ready()):
			print("pm_query not is_ready")
			return None

		cmd_str = rquery['cmd']
		if cmd_str in ['GREP']:
			return QueryBase.query(self, rquery)
		req     = rquery['req']
		opt     = rquery['opt']
		if opt == None:
			opt = []

		pargs = [idutils_lid_cmd, '-R', 'grep']
		if cmd_str == 'FIL':
			pargs = ['fnid', '-S', 'newline']
		#elif cmd_str == 'TXT':
			#pargs += ['-l']
		elif cmd_str in ['-->', '<--']:
			pargs += ['-l']

		if 'substring' in opt:
			if cmd_str == 'FIL':
				req = '*' + req + '*'
			else:
				req = '.*' + req + '.*'
				#pargs += ' -s'
		if 'ignorecase' in opt:
			if cmd_str == 'FIL':
				pass
			else:
				pargs += ['-i']

		pargs += [ '--', req ]
		qsig = IdProcess(self.conf.c_dir, [cmd_str, req]).run_query_process(pargs, req, rquery)
		return qsig

	def rebuild(self):
		if (not self.conf.is_ready()):
			print("pm_query not is_ready")
			return None
		pargs = os.getenv('SEASCOPE_IDUTILS_MKID_CMD', '').strip().split()
		if not len(pargs):
			pargs = [ 'mkid', '-s' ]
		qsig = IdProcess(self.conf.c_dir, None).run_rebuild_process(pargs)
		return qsig

	def query_fl(self):
		if not os.path.exists(os.path.join(self.conf.c_dir, 'ID')):
			return []
		pargs = [ 'fnid', '-S', 'newline', '-f', 'ID' ]
		qsig = IdProcess(self.conf.c_dir, None).run_query_fl(pargs)
		return qsig

	def id_is_open(self):
		return self.conf != None
	def id_is_ready(self):
		return self.conf.is_ready()



# Copyright (c) 2010 Anil Kumar
# All rights reserved.
#
# License: BSD 

import subprocess
import re, os

def _eintr_retry_call(func, *args):
	while True:
		try:
			return func(*args)
		except OSError as e:
			if e.errno == errno.EINTR:
				continue
			raise

def cmdForFile(f):
	#args = 'ctags -n -u --fields=+K -f - --extra=+q'
	#args = 'ctags -n -u --fields=+Ki -f -'
	args = 'ctags -n -u --fields=+K -f -'

	opt_I_file = os.getenv('SEASCOPE_CTAGS_OPT_I_FILE')
	if opt_I_file and os.path.isfile(opt_I_file):
		args += ' -I ' + opt_I_file

	suffix_cmd_map = []
	custom_map = os.getenv('SEASCOPE_CTAGS_SUFFIX_CMD_MAP')
	if custom_map:
		custom_map = eval(custom_map)
		suffix_cmd_map += custom_map
	suffix_cmd_map.append( ['', args] )
	for (suffix, cmd) in suffix_cmd_map:
		if f.endswith(suffix):
			return cmd
	return None

def ct_query(filename):
	args = cmdForFile(filename)
	args = args.split()
	args.append(filename)
	try:
		proc = subprocess.Popen(args, stdout=subprocess.PIPE)
		(out_data, err_data) = _eintr_retry_call(proc.communicate)
		err_data = err_data.decode() if err_data else ''
		out_data = out_data.decode().split('\n')
	except Exception as e:
		out_data =  [
				'Failed to run ctags cmd\tignore\t0;\t ',
				'cmd: %s\tignore\t0;\t ' % ' '.join(args),
				'error: %s\tignore\t0;\t ' % str(e),
				'ctags not installed ?\tignore\t0;\t ',
			]
	res = []
	for line in out_data:
		if (line == ''):
			break
		line = line.split('\t')
		num = line[2].split(';', 1)[0]
		line = [line[0], num, line[3]]
		res.append(line)
	return res

is_OrderedDict_available = False
try:
	# OrderedDict available only in python >= 2.7
	from collections import OrderedDict
	is_OrderedDict_available = True
except:
	pass

def emptyOrderedDict():
	if is_OrderedDict_available:
		return OrderedDict({})
	return {}

class CtagsTreeBuilder:
	def __init__(self):
		self.symTree = emptyOrderedDict()

	def cmdForFile(self, f):
		#args = 'ctags -n -u --fields=+K -f - --extra=+q'
		#args = 'ctags -n -u --fields=+Ki -f -'
		args = 'ctags -n -u --fields=+K-f-t -f -'

		opt_I_file = os.getenv('SEASCOPE_CTAGS_OPT_I_FILE')
		if opt_I_file and os.path.isfile(opt_I_file):
			args += ' -I ' + opt_I_file

		suffix_cmd_map = []
		custom_map = os.getenv('SEASCOPE_CTAGS_SUFFIX_CMD_MAP')
		if custom_map:
			custom_map = eval(custom_map)
			suffix_cmd_map += custom_map
		suffix_cmd_map.append( ['', args] )
		for (suffix, cmd) in suffix_cmd_map:
			if f.endswith(suffix):
				return cmd
		return None

	def runCtags(self, f):
		args = self.cmdForFile(f)
		args = args.split()
		args.append(f)
		# In python >= 2.7 can use subprocess.check_output
		# output = subprocess.check_output(args)
		# return output
		proc = subprocess.Popen(args, stdout=subprocess.PIPE)
		(out_data, err_data) = proc.communicate()
		out_data = out_data.decode()
		return out_data

	def parseCtagsOutput(self, data):
		data = re.split(r'\r?\n', data)
		res = []
		for line in data:
			if line == '':
				continue
			try:
				line = line.split('\t', 4)
				res.append(line)
			except:
				print('bad line:', line)
		return res


	def addToSymLayout(self, sc):
		t = self.symTree
		if sc and sc != '':
			for s in re.split(r'::|\.', sc):
				if s not in t:
					t[s] = emptyOrderedDict()
				t = t[s]

	def addToSymTree(self, sc, line):
		t = self.symTree
		if sc and sc != '':
			for s in re.split(r'::|\.', sc):
				assert s in t
				t = t[s]

		cline = [line[0], line[2].split(';')[0], line[3]]
		if line[0] in t:
			#print line[0], 'in', t
			x = t[line[0]]
			if '+' not in x:
				x['+'] = cline
				return
		if '*' not in t:
			t['*'] = []
		t['*'].append(cline)
		#print '...', t, line

	def buildTree(self, data):
		type_list = [ 'namespace', 'class', 'interface', 'struct', 'union', 'enum', 'function' ]
		# build layout using 5th field
		for line in data:
			if len(line) == 4:
				continue
			try:
				sd = dict([ x.split(':', 1) for x in line[4].split('\t')])
			except:
				print('bad line', line)
				continue
			line[4] = sd
			count = 0
			for t in type_list:
				if t in sd:
					self.addToSymLayout(sd[t])
					count = count + 1
			if count != 1:
				print('******** count == 1 *********')
				print(data)
				print(line)
			#assert count == 1
		
		if len(self.symTree) == 0:
			return (data, False)
		
		for line in data:
			if len(line) == 4:
				self.addToSymTree(None, line)
				continue
			sd = line[4]
			count = 0
			for t in type_list:
				if t in sd:
					self.addToSymTree(sd[t], line)
					count = count + 1
			if count != 1:
				print('******** count == 1 *********')
				print(data)
				print(line)
			#assert count == 1

		return (self.symTree, True)

	def doQuery(self, filename):
		try:
			output = self.runCtags(filename)
			output = self.parseCtagsOutput(output)
			output = self.buildTree(output)
		except Exception as e:
			print(str(e))
			output = [None, False]
		return output


def ct_tree_query(filename):
	ct = CtagsTreeBuilder()
	output = ct.doQuery(filename)
	return output

if __name__ == '__main__':
	import optparse
	import sys
	depth = 0
	def recursePrint(t):
		global depth
		for k, v in list(t.items()):
			if k == '*':
				for line in v:
					print('%s%s' % (' ' * depth, line))
				continue
			if k == '+':
				continue

			if '+' in v:
				k = v['+']
			print('%s%s' % (' ' * depth, k))
				
			depth = depth + 4
			recursePrint(v)
			depth = depth - 4

	op = optparse.OptionParser()
	(options, args) = op.parse_args()
	if len(args) != 1:
		print('Please specify a file')
		sys.exit(-1)

	(output, isTree) = ct_tree_query(args[0])
	if isTree:
		recursePrint(output)
	else:
		for line in output:
			print(line)

